from enum import Enum


class RecordingType(Enum):
    """
    Various sensor measurement types

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    BINARY = "Binary"
    LINEAR = "Linear"
    CONTINUOUS = "Continuous"
    EXPONENTIAL = "Exponential"



from metamenth.subsystem.interfaces.abstract_subsystem import AbstractSubsystem
from metamenth.enumerations import ApplianceCategory
from metamenth.enumerations import ApplianceType
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.datatypes.rated_device_measure import RatedDeviceMeasure
from metamenth.datatypes.continuous_measure import ContinuousMeasure
from metamenth.datatypes.interfaces.abstract_dynamic_entity import AbstractDynamicEntity


class Appliance(AbstractSubsystem, AbstractDynamicEntity):
    def __init__(self, name: str, appliance_category: [ApplianceCategory],
                 appliance_type: ApplianceType, manufacturer: str = None,
                 consumption_capacity: AbstractMeasure = None,
                 rated_device_measure: RatedDeviceMeasure = None,
                 operating_conditions: [ContinuousMeasure] = None):
        """
        Defines an appliance in a built environment
        :param name: the unique name of the appliance
        :param appliance_category: the appliance category e.g., home appliance.
        :param appliance_type: the appliance type, e.g., photocopier
        :param manufacturer: the manufacturer of the appliance
        :param consumption_capacity: the power consumption capacity of the appliance
        :param rated_device_measure: the rated device measure of the appliance
        """
        AbstractSubsystem.__init__(self, name)
        AbstractDynamicEntity.__init__(self)

        self._appliance_type = None
        self._manufacturer = manufacturer
        self._consumption_capacity = consumption_capacity
        self._rated_device_measure = rated_device_measure
        self._appliance_category: [ApplianceCategory] = []
        self._operating_conditions: [ContinuousMeasure] = operating_conditions

        self.appliance_type = appliance_type
        self.appliance_category = appliance_category

    @property
    def appliance_type(self) -> ApplianceType:
        return self._appliance_type

    @appliance_type.setter
    def appliance_type(self, value: ApplianceType):
        if value is not None:
            self._appliance_type = value
        else:
            raise ValueError("appliance_type must be of type ApplianceType")

    @property
    def manufacturer(self) -> str:
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value: str):
        self._manufacturer = value

    @property
    def consumption_capacity(self) -> AbstractMeasure:
        return self._consumption_capacity

    @consumption_capacity.setter
    def consumption_capacity(self, value: AbstractMeasure):
        self._consumption_capacity = value

    @property
    def rated_device_measure(self) -> RatedDeviceMeasure:
        return self._rated_device_measure

    @rated_device_measure.setter
    def rated_device_measure(self, value: RatedDeviceMeasure):
        self._rated_device_measure = value

    @property
    def appliance_category(self) -> [ApplianceCategory]:
        return self._appliance_category.copy() if self._appliance_category else []

    @appliance_category.setter
    def appliance_category(self, value: [ApplianceCategory]):
        if value is not None and type(value) is list:
            self._appliance_category.extend(value)
        else:
            raise ValueError("appliance_category should be a list of type ApplianceCategory")

    @property
    def operating_conditions(self) -> [ContinuousMeasure]:
        return self._operating_conditions.copy() if self._operating_conditions else []

    @operating_conditions.setter
    def operating_conditions(self, value: [ContinuousMeasure]):
        if value is not None and type(value) is list:
            self._operating_conditions.extend(value)

    def __str__(self):
        return (
            f"Appliance ({super().__str__()}"
            f"Manufacturer: {self.manufacturer}, "
            f"Appliance Type: {self.appliance_type.value}, "
            f"Appliance Category: {self.appliance_category}, "
            f"Consumption Capacity: {self.consumption_capacity}, "
            f"Operating Conditions: {self.operating_conditions}, "
            f"Rated Device Measure: {self.rated_device_measure})"
        )




from metamenth.structure.interfaces.abstract_floor_space import AbstractFloorSpace
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.enumerations import OpenSpaceType


class OpenSpace(AbstractFloorSpace):
    """
    Defines an open space on a floor of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, name: str, area: AbstractMeasure, space_type: OpenSpaceType, location: str = None):
        """
        :param area: The area of the open space.
        :param location: The location of the open space (three words delimited with a period).
        :param space_type: The type of open space (enum).
        """
        super().__init__(area, name, location)
        self._space_type = None

        # Apply validation
        self.space_type = space_type

    @property
    def space_type(self) -> OpenSpaceType:
        return self._space_type

    @space_type.setter
    def space_type(self, value: OpenSpaceType):
        if value is None:
            raise ValueError('space_type must be of type OpenSpaceType')
        self._space_type = value

    def accept(self, visitor):
        """
        visitor method to accept
        visit operation for the current open space
        :param visitor: the visitor object
        """
        visitor.visit_open_space(self)

    def __str__(self):
        return f"OpenSpace ({super().__str__()} Space Type: {self.space_type.value})"



from metamenth.enumerations.abstract_enum import AbstractEnum


class CompressorType(AbstractEnum):
    """
    Types of compressor

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    RECIPROCATING_HERMETIC = "ReciprocatingHermetic"
    RECIPROCATING_SEMI_HERMETIC = "ReciprocatingSemiHermetic"
    RECIPROCATING_OPEN = "ReciprocatingOpen"
    ROTARY_VANE = "RotaryVane"
    SCROLL = "Scroll"
    SCREW = "Screw"
    CENTRIFUGAL = "Centrifugal"
    OTHER = "Other"




from uuid import uuid4
from metamenth.enumerations import BuildingType
from metamenth.datatypes.address import Address
from typing import List
from .floor import Floor
from .room import Room
from .open_space import OpenSpace
from .envelope import Envelope
from typing import Optional
from metamenth.measure_instruments.weather_station import WeatherStation
from metamenth.measure_instruments.meter import Meter
from metamenth.measure_instruments.interfaces.abstract_reader import AbstractReader
from metamenth.utils import EntityRemover
from metamenth.utils import EntityInsert
from metamenth.enumerations import BuildingEntity
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.virtual.zone import Zone
from metamenth.utils import StructureSearch
from typing import Dict
from metamenth.enumerations import RoomType
from metamenth.enumerations import OpenSpaceType
from metamenth.enumerations import TerrainType
from metamenth.enumerations import SolarDistributionType
from metamenth.observers.observable import Observable
from metamenth.misc import StateTrackDecorator
from metamenth.utils import StructureEntitySearch
from metamenth.enumerations import MeterType
from metamenth.subsystem.building_control_system import BuildingControlSystem
from metamenth.datatypes.schedulable_entity import SchedulableEntity


class Building(Observable):
    """
    A representation of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, construction_year: int, height: AbstractMeasure, floor_area: AbstractMeasure,
                 internal_mass: AbstractMeasure, address: Address, building_type: BuildingType, floors: List[Floor]):
        """
        :param construction_year: The construction year of the building
        :param height: The height of the building
        :param floor_area: The floor area of the building
        :param internal_mass: The internal mass of the building
        :param address: The address of the building
        :param building_type: The type of building
        """
        super().__init__()
        self._UID = str(uuid4())
        self._name = None
        self._construction_year = None
        self._height = None
        self._floor_area = None
        self._internal_mass = None
        self._address = None
        self._building_type = None
        self._terrain = None
        self._solar_distribution = None
        self._schedulable_entity = SchedulableEntity()
        self._envelope: Envelope = Optional[None]
        self._floors = []
        self._meters: [AbstractReader] = []
        self._weather_stations: List[WeatherStation] = []
        self._zones: List[Zone] = []
        self._control_systems: [BuildingControlSystem] = []
        self.track_state = False

        # apply validation
        self.construction_year = construction_year
        self.height = height
        self.floor_area = floor_area
        self.internal_mass = internal_mass
        self.address = address
        self.building_type = building_type
        self.add_floors(floors)

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def construction_year(self) -> int:
        return self._construction_year

    @construction_year.setter
    def construction_year(self, value):
        if value is not None:
            self._construction_year = value
        else:
            raise ValueError("construction_year must be a number")

    @property
    def height(self) -> AbstractMeasure:
        return self._height

    @height.setter
    def height(self, value: AbstractMeasure):
        if value is not None:
            self._height = value
        else:
            raise ValueError("height must be of type AbstractMeasure")

    @property
    def floor_area(self) -> AbstractMeasure:
        return self._floor_area

    @floor_area.setter
    @StateTrackDecorator
    def floor_area(self, value: AbstractMeasure):
        if value is not None:
            # write state before updating
            self._floor_area = value
        else:
            raise ValueError("floor_area must be of type AbstractMeasure")

    @property
    def internal_mass(self) -> AbstractMeasure:
        return self._internal_mass

    @internal_mass.setter
    def internal_mass(self, value: AbstractMeasure):
        if value is not None:
            self._internal_mass = value
        else:
            raise ValueError("internal_mass must be of type AbstractMeasure")

    @property
    def address(self) -> Address:
        return self._address

    @address.setter
    @StateTrackDecorator
    def address(self, value: Address):
        if value is not None:
            self._address = value
        else:
            raise ValueError("address must be of type Address")

    @property
    def building_type(self) -> BuildingType:
        return self._building_type

    @building_type.setter
    @StateTrackDecorator
    def building_type(self, value: BuildingType):
        if value is not None:
            self._building_type = value
        else:
            raise ValueError("building_type must be of type BuildingType")

    @property
    def floors(self):
        raise AttributeError("Cannot get floors")

    @floors.setter
    def floors(self, value: List[Floor]):
        if value is not None and len(value) > 0:
            self._floors = value
        else:
            raise ValueError("floors must be of type [Floor] and must not be empty")

    @property
    def weather_stations(self):
        raise AttributeError("Cannot get weather stations")

    @weather_stations.setter
    def weather_stations(self, value: List[WeatherStation]):
        if value is not None:
            self._weather_stations = value
        else:
            raise ValueError("weather_stations must be of type [WeatherStation]")

    @property
    def meters(self):
        raise AttributeError("Cannot get meters")

    @meters.setter
    def meters(self, value: List[AbstractReader]):
        if value is not None:
            self._meters = value
        else:
            raise ValueError("meters must be of type [Meter]")

    @property
    def envelope(self) -> Envelope:
        return self._envelope

    @envelope.setter
    @StateTrackDecorator
    def envelope(self, value: Envelope):
        if value is not None:
            self._envelope = value
        else:
            raise ValueError("envelope must be of type Envelope")

    @property
    def schedules(self):
        raise AttributeError("Cannot get schedules")

    @property
    def zones(self):
        return self._zones

    @zones.setter
    def zones(self, value):
        if value is not None:
            self._zones = value
        else:
            raise ValueError('zones must be of type [Zone]')

    @property
    def control_systems(self):
        return self._control_systems.copy()

    def add_control_system(self, control_system: BuildingControlSystem):
        if control_system:
            self._control_systems.append(control_system)

    @property
    def schedulable_entity(self) -> SchedulableEntity:
        return self._schedulable_entity

    @schedulable_entity.setter
    def schedulable_entity(self, value: SchedulableEntity):
        if value is None:
            raise ValueError("schedules should be of type SchedulableEntity")
        self._schedulable_entity = value

    @property
    def terrain(self) -> TerrainType:
        return self._terrain

    @terrain.setter
    def terrain(self, value: TerrainType):
        if value is not None:
            self._terrain = value
        else:
            raise ValueError("terrain must be of type TerrainType")

    @property
    def solar_distribution(self) -> SolarDistributionType:
        return self._solar_distribution

    @solar_distribution.setter
    def solar_distribution(self, value: SolarDistributionType):
        if value is not None:
            self._solar_distribution = value
        else:
            raise ValueError("solar_distribution must be of type SolarDistributionType")

    @StateTrackDecorator
    def add_weather_station(self, weather_station: WeatherStation):
        """
        Adds a weather station to a building
        :param weather_station: a station to measure various weather elements
        :return:
        """
        EntityInsert.insert_building_entity(self._weather_stations, weather_station)
        return self

    @StateTrackDecorator
    def remove_weather_station(self, weather_station: WeatherStation):
        """
        Adds a weather station to a building
        :param weather_station: a station to measure various weather elements
        :return:
        """
        EntityRemover.remove_building_entity(self._weather_stations, weather_station)

    def add_meter(self, meter: AbstractReader):
        """
        Adds a meter to a building
        :param meter: a meter to measure some phenomena e.g. energy consumption
        :return:
        """
        EntityInsert.insert_building_entity(self._meters, meter)
        return self

    def remove_meter(self, meter: AbstractReader):
        """
        Adds a meter to a building
        :param meter: a meter to measure some phenomena e.g. energy consumption
        :return:
        """
        EntityRemover.remove_building_entity(self._meters, meter)

    def add_floors(self, floors: List[Floor]):
        """
        Add multiple unique (by floor number) floors to a building
        :param floors: the floors to add to this building
        :return:
        """
        EntityInsert.insert_building_entity(self._floors, floors, BuildingEntity.FLOOR_SPACE.value)
        return self  # necessary for method chaining

    def remove_floor(self, floor: Floor):
        """
        Removes floor from a building
        :param floor: the floor to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._floors, floor)

    def get_floor_by_uid(self, uid: str) -> Floor:
        """
        Retrieves a floor given the uid
        :param uid: the uid of the floor
        :return:
        """
        return StructureSearch.search_by_id(self._floors, uid)

    def get_floor_by_number(self, floor_number: int) -> Floor:
        """
        Retrieves a floor given the floor number
        :param floor_number: the number assigned to the floor
        :return:
        """
        return StructureSearch.search_by_number(self._floors, floor_number)

    def get_floors(self, search_term: Dict = None) -> List[Floor]:
        """
        Retrieves floors given the attributes and their values
        :param search_term: the uid of the floor
        :return:
        """
        return StructureSearch.search(self._floors, search_term)

    def get_weather_station_by_name(self, name: str) -> WeatherStation:
        """
        Returns a weather station
        :param name: the name of the weather station
        :return:
        """
        return StructureEntitySearch.search_by_name(self._weather_stations, name)

    def get_weather_station_by_uid(self, uid: str) -> WeatherStation:
        """
        Returns a weather station
        :param uid: the unique identifier of the weather station
        :return:
        """
        return StructureEntitySearch.search_by_id(self._weather_stations, uid)

    def get_weather_stations(self, search_term: Dict = None) -> [WeatherStation]:
        """
        Returns a list of weather stations
        :param search_term: attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._weather_stations, search_term)

    def get_meter_by_uid(self, uid: str) -> AbstractReader:
        """
        Returns a meter based on uid
        :param uid: the uid of the meter
        :return:
        """
        return StructureEntitySearch.search_by_id(self._meters, uid)

    def get_meter_by_type(self, meter_type: MeterType) -> [Meter]:
        """
        Returns a meter based on type of meter
        :param meter_type: the type of meter
        :return:
        """
        return StructureEntitySearch.search(self._meters, {'meter_type': meter_type})

    def get_meters(self, search_terms: Dict = None) -> [AbstractReader]:
        """
        Returns a meter based on some attributes and their values
        :param search_terms: attributes and value key pairs
        :return:
        """
        return StructureEntitySearch.search(self._meters, search_terms)

    def get_zone_by_name(self, name) -> Zone:
        """
        Search zones by name
        :param name:  the name of the zone
        :return:
        """
        return StructureEntitySearch.search_by_name(self._zones, name)

    def get_zone_by_uid(self, uid) -> Zone:
        """
        Search zones by uid
        :param uid: the unique identifier of the overlapping zone
        :return:
        """
        return StructureEntitySearch.search_by_id(self._zones, uid)

    def get_zones(self, search_terms: Dict = None) -> [Zone]:
        """
        Search zones by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._zones, search_terms)

    @StateTrackDecorator
    def add_room(self, floor_uid: str, name: str, area: AbstractMeasure,
                 room_type: RoomType,
                 location: str = None):
        """
        Adds a room to a building's floor
        :param floor_uid: the unique ID of the floor
        :param area: the area of the room
        :param name: the name of the room
        :param room_type: the tupe of room
        :param location: the location of the room
        :return:
        """
        if not self.get_floor_by_uid(floor_uid):
            raise ValueError("Cannot add room without a floor")
        self.get_floor_by_uid(floor_uid).add_rooms([Room(area, name, room_type, location)])
        return self  # Return self for method chaining

    @StateTrackDecorator
    def add_open_space(self, floor_uid: str, name: str, area: AbstractMeasure,
                       space_type: OpenSpaceType, location: str = None):
        """
        Adds an open space to a building's floor
        :param floor_uid: the floor UID
        :param name: the name of the open space
        :param area: the area of the open space
        :param space_type: the type of open space
        :param location: the location of the open space
        :return:
        """
        if not self.get_floor_by_uid(floor_uid):
            raise ValueError("Cannot add open space without a floor")
        self.get_floor_by_uid(floor_uid).add_open_spaces([OpenSpace(name, area, space_type, location)])
        return self  # Return self for method chaining

    def accept(self, visitor):
        """
        visitor method to accept
        visit operation to building floors
        :param visitor: the visitor object
        """
        visitor.visit_building(self)

    def __str__(self):
        floors_info = "\n".join([f"  - Floor {floor.number}: {floor}" for floor in self._floors])
        weather_stations_info = "\n".join([f"  - {station}" for station in self._weather_stations])
        schedules_info = "\n".join([f"  - {schedule}" for schedule in self._schedulable_entity.get_schedules()])
        meter_info = "\n".join([f"  - {meter}" for meter in self._meters])

        return (f"Building("
                f"UID: {self.UID}, "
                f"Construction Year: {self.construction_year}, "
                f"Height: {self.height}, "
                f"Floor Area: {self.floor_area}, "
                f"Internal Mass: {self.internal_mass}, "
                f"Address: {self.address}, "
                f"Building Type: {self.building_type}, "
                f"Terrain: {self.terrain}, "
                f"Solar Distribution: {self.solar_distribution}, "
                f"Floor Count: {len(self._floors)}, "
                f"Weather Stations Count: {len(self._weather_stations)}, "
                f"Schedules: {len(self._schedulable_entity.get_schedules())}, "
                f"Floors:\n{floors_info}, "
                f"Weather Stations:\n{weather_stations_info}, "
                f"Schedules:\n{schedules_info}, "
                f"Meters:\n{meter_info}, "
                f"Envelope: {self.envelope}")



from metamenth.subsystem.interfaces.abstract_ventilation_component import AbstractVentilationComponent
from metamenth.enumerations import HeatingType
from metamenth.enumerations import PowerState


class BaseboardHeater(AbstractVentilationComponent):

    def __init__(self, name: str, heating_type: HeatingType, power_state: PowerState):
        super().__init__(name)
        self._heating_type = None
        self._power_state = None

        self.heating_type = heating_type
        self.power_state = power_state

    @property
    def heating_type(self) -> HeatingType:
        return self._heating_type

    @heating_type.setter
    def heating_type(self, value: HeatingType):
        if value is not None:
            self._heating_type = value
        else:
            raise ValueError("heating_type must be of type HeatingType")

    @property
    def power_state(self) -> PowerState:
        return self._power_state

    @power_state.setter
    def power_state(self, value: PowerState):
        if value is not None:
            self._power_state = value
        else:
            raise ValueError("power_state must be of type PowerState")

    def __str__(self):
        return (
            f"BaseboardHeater ({super().__str__()}"
            f"Heating Type: {self.heating_type}"
            f"Power State: {self.power_state.value})"
        )




from metamenth.energysystem.interfaces.abstract_energy_system import AbstractEnergySystem
from metamenth.enumerations import MeasurementUnit


class RenewableEnergySystem(AbstractEnergySystem):
    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit):
        super().__init__(name, inverter, unit)




from metamenth.energysystem.non_renewable_energy_system import NonRenewableEnergySystem
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import EngineType
from metamenth.enumerations import EngineMode
from metamenth.enumerations import EngineSubType
from metamenth.misc import Validate


class Engine(NonRenewableEnergySystem):

    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit,
                 engine_type: EngineType, engine_sub_type: EngineSubType, engine_mode: EngineMode):
        super().__init__(name, inverter, unit)
        self._engine_type = None
        self._engine_mode = None
        self._engine_sub_type = None

        self.engine_type = engine_type
        self.engine_mode = engine_mode
        self.engine_sub_type = engine_sub_type

    @property
    def engine_type(self) -> EngineType:
        return self._engine_type

    @engine_type.setter
    def engine_type(self, value: EngineType):
        if value is None:
            raise ValueError("engine_type should be of type EngineType")
        self._engine_type = value

    @property
    def engine_sub_type(self) -> EngineSubType:
        return self._engine_sub_type

    @engine_sub_type.setter
    def engine_sub_type(self, value: EngineSubType):
        if value is None:
            raise ValueError("engine_sub_type should be of type EngineSubType")
        else:
            if Validate.validate_engine_fuel(self.engine_type.value, value.value):
                self._engine_sub_type = value
            else:
                raise ValueError(f'{value} is an invalid value for {self.engine_type}')

    @property
    def engine_mode(self) -> EngineMode:
        return self._engine_mode

    @engine_mode.setter
    def engine_mode(self, value: EngineMode):
        if value is None:
            raise ValueError("engine_mode should be of type EngineMode")
        self._engine_mode = value

    def __str__(self):
        return (
            f"Engine("
            f"{super().__str__()}, "
            f"Engine Type: {self.engine_type.value}, "
            f"Engine SubType: {self.engine_sub_type.value}, "
            f"Engine Mode: {self.engine_mode.value})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class RoomType(AbstractEnum):
    """
    Different types of rooms on a floor of a building.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    OFFICE = "Office"
    KITCHEN = "Kitchen"
    BEDROOM = "Bedroom"
    LIVING = "Living"
    DINNING = "Dinning"
    PANTRY = "Pantry"
    STORAGE = "Storage"
    LIBRARY = "Library"
    WASHROOM = "Washroom"
    RESEARCH_LAB = "ResearchLab"
    COMPUTER_LAB = "ComputerLab"
    LAUNDRY = "Laundry"
    GARAGE = "Garage"
    INDOOR_PARKING = "IndoorParking"
    DATA_CENTER = "DataCenter"
    MECHANICAL = "Mechanical"
    CLASSROOM = "Classroom"
    COMMON = "Common"
    LAVATORY = "Lavatory"
    HOUSE_KEEPING = "HouseKeeping"
    MAINTENANCE = "Maintenance"
    CAFETERIA = "Cafeteria"
    FILE = "File"
    LOUNGE = "Lounge"
    SUPPORT_SPACE = "SupportSpace"
    STAFF = "Staff"
    ENTERTAINMENT = "Entertainment"
    CONCIERGE = "Concierge"
    FIRE = "Fire"
    ELECTRIC = "Electric"
    MULTIFUNCTIONAL = "Multifunctional"
    TRIAGE = "Triage"
    LOCKER = "Locker"
    JANITOR_CLOSET = "JanitorCloset"
    PHOTO = "Photo"
    READING = "Reading"
    COMPUTER_SERVICE = "ComputerServices"
    COMMUNITY_SERVICE = "CommunityServices"
    CONSULTATION = "Consultation"
    SEMINAR = "Seminar"
    SPECIALISED_CLASSROOM = "SpecialisedClassroom"
    MAIL_ROOM = "MailRoom"
    BOOKSTORE = "BookStore"
    STUDY_ROOM = "StudyRoom"
    OTHER = "Other"




from metamenth.enumerations.abstract_enum import AbstractEnum


class FCUPipeSystem(AbstractEnum):
    """
    Piping system for FCUs

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    TWO_PIPE = "TwoPipe"
    FOUR_PIPE = "FourPipe"





from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.enumerations import FilterType
from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.misc import Validate


class Filter(AbstractHVACComponent):
    def __init__(self, name: str, filter_type: FilterType, air_flow_rate: BinaryMeasure = None,
                 life_span: int = None, efficiency: int = None):
        """
        Models a filter in hvac systems
        :param name: the unique name of the heat exchanger
        :param filter_type: the type of filter
        :param air_flow_rate: the rate of air flow usually measured in cubic feet per meter
        :param life_span: the life span in days, of the filter
        :param efficiency: a number from 1 to 16 indicating how efficient the filter is
        """
        super().__init__(name)
        self._filter_type = None
        self._air_flow_rate = air_flow_rate
        self._life_span = None
        self._efficiency = None

        self.filter_type = filter_type
        self.life_span = life_span
        self.efficiency = efficiency

    @property
    def filter_type(self) -> FilterType:
        return self._filter_type

    @filter_type.setter
    def filter_type(self, value: FilterType):
        if value is not None:
            self._filter_type = value
        else:
            raise ValueError("filter_type must be of type FilterType")

    @property
    def air_flow_rate(self) -> BinaryMeasure:
        return self._air_flow_rate

    @air_flow_rate.setter
    def air_flow_rate(self, value: BinaryMeasure):
        self._air_flow_rate = value

    @property
    def life_span(self) -> int:
        return self._life_span

    @life_span.setter
    def life_span(self, value: int):
        if value:
            if value > 0:
                self._life_span = value
            else:
                raise ValueError('life_span must be a positive integer')

    @property
    def efficiency(self) -> int:
        return self._efficiency

    @efficiency.setter
    def efficiency(self, value: int):
        self._efficiency = Validate.validate_number_range(value, (1, 16))

    def __str__(self):
        return (
            f"Filter ({super().__str__()}"
            f"Filter Type: {self.filter_type}, "
            f"Lifespan: {self.life_span}, "
            f"Air Flow Rate: {self.air_flow_rate}, "
            f"Efficiency: {self.efficiency})"
        )



from uuid import uuid4
from metamenth.enumerations import DuctType
from metamenth.enumerations import DuctSubType
from metamenth.datatypes.interfaces.abstract_dynamic_entity import AbstractDynamicEntity
from metamenth.virtual.interfaces.abstract_zonal_entity import AbstractZonalEntity
from metamenth.subsystem.hvac_components.damper import Damper
from metamenth.subsystem.hvac_components.fan import Fan
from metamenth.subsystem.hvac_components.heat_exchanger import HeatExchanger
from typing import List
from metamenth.subsystem.hvac_components.duct_connection import DuctConnection
from metamenth.utils import EntityInsert
from metamenth.utils import EntityRemover
from metamenth.enumerations import BuildingEntity
from typing import Dict
from metamenth.utils import StructureEntitySearch
from metamenth.subsystem.hvac_components.air_volume_box import AirVolumeBox
from metamenth.subsystem.hvac_components.filter import Filter


class Duct(AbstractDynamicEntity, AbstractZonalEntity):

    def __init__(self, name: str, duct_type: DuctType):
        AbstractDynamicEntity.__init__(self)
        AbstractZonalEntity.__init__(self)
        self._UID = str(uuid4())
        self._name = None
        self._duct_type = None
        self._duct_sub_type = None
        self._connections = None
        self._heat_exchangers: List[HeatExchanger] = []
        self._fans: List[Fan] = []
        self._dampers: List[Damper] = []
        self._connected_air_volume_box: [AirVolumeBox] = []
        self._filters: [Filter] = []
        self.name = name
        self.duct_type = duct_type

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is not None:
            self._name = value
        else:
            raise ValueError("name should be of type string")

    @property
    def duct_type(self) -> DuctType:
        return self._duct_type

    @duct_type.setter
    def duct_type(self, value: DuctType):
        if value is not None:
            self._duct_type = value
        else:
            raise ValueError("duct_type should be of type DuctType")

    @property
    def duct_sub_type(self) -> DuctSubType:
        return self._duct_sub_type

    @duct_sub_type.setter
    def duct_sub_type(self, value: DuctSubType):
        self._duct_sub_type = value

    @property
    def connections(self) -> DuctConnection:
        return self._connections

    @connections.setter
    def connections(self, value: DuctConnection):
        if value is not None:
            self._connections = value
        else:
            raise ValueError("connections should be of type DuctConnection")

    def add_heat_exchanger(self, new_heat_exchanger: HeatExchanger):
        """
        Adds heat exchangers
        :param new_heat_exchanger: a heat exchanger to be added to this duct
        :return:
        """
        EntityInsert.insert_building_entity(self._heat_exchangers, new_heat_exchanger,
                                            BuildingEntity.HVAC_COMPONENT.value)

    def add_filter(self, new_filter: Filter):
        """
        Adds filters to duct
        :param new_filter: a heat exchanger to be added to this duct
        :return:
        """
        EntityInsert.insert_building_entity(self._filters, new_filter, BuildingEntity.HVAC_COMPONENT.value)

    def remove_filter(self, hvac_filter: Filter):
        """
        Removes a filter from a duct
        :param hvac_filter: the filter to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._filters, hvac_filter)

    def get_filters(self, search_terms: Dict = None) -> [Filter]:
        """
        Search filters by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._filters, search_terms)

    def add_fan(self, new_fan: Fan):
        """
        Adds fans
        :param new_fan: a fan to be added to this duct
        :return:
        """
        EntityInsert.insert_building_entity(self._fans, new_fan, BuildingEntity.HVAC_COMPONENT.value)

    def add_connected_air_volume_box(self, new_vav_box: AirVolumeBox):
        """
        Adds VAV boxes connected to this duct
        :param new_vav_box: a fan to be added to this duct
        :return:
        """
        EntityInsert.insert_building_entity(self._connected_air_volume_box, new_vav_box,
                                            BuildingEntity.HVAC_COMPONENT.value)

    def remove_connected_air_volume_box(self, vav_box: AirVolumeBox):
        """
        Removes a VAV box from a duct
        :param vav_box: the VAV box to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._connected_air_volume_box, vav_box)

    def add_damper(self, new_damper: Damper):
        """
        Adds dampers
        :param new_damper: a damper to be added to this duct
        :return:
        """
        EntityInsert.insert_building_entity(self._dampers, new_damper, BuildingEntity.HVAC_COMPONENT.value)

    def remove_fan(self, fan: Fan):
        """
        Removes a fan from a duct
        :param fan: the fan to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._fans, fan)

    def remove_heat_exchanger(self, heat_exchanger: HeatExchanger):
        """
        Removes a heat exchanger from a duct
        :param heat_exchanger: the heat exchanger to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._heat_exchangers, heat_exchanger)

    def remove_damper(self, damper: Damper):
        """
        Removes a damper from a duct
        :param damper: the fan to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._dampers, damper)

    def get_heat_exchangers(self, search_terms: Dict = None) -> [HeatExchanger]:
        """
        Search heat exchangers by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._heat_exchangers, search_terms)

    def get_dampers(self, search_terms: Dict = None) -> [Damper]:
        """
        Search dampers by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._dampers, search_terms)

    def get_fans(self, search_terms: Dict = None) -> [Fan]:
        """
        Search fans by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._fans, search_terms)

    def get_connected_air_volume_boxes(self, search_terms: Dict = None) -> [AirVolumeBox]:
        """
        Search air volume boxes by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._connected_air_volume_box, search_terms)

    def __eq__(self, other):
        # ducts are equal if they share the same name
        if isinstance(other, Duct):
            # Check for equality based on the name and UID attribute
            return self.name == other.name and self.UID == other.UID
        return False

    def __str__(self):
        return (
            f"Duct("
            f"UID: {self.UID}, "
            f"name: {self.name}, "
            f"Type: {self.duct_type}, "
            f"SubType: {self.duct_sub_type}, "
            f"Fans: {self._fans}, "
            f"Heat Exchangers: {self._heat_exchangers}, "
            f"Dampers: {self._dampers}, "
            f"Air Volume Box: {self._connected_air_volume_box}, "
            f"Filters: {self._filters}, "
            f"Connection: {self.connections})"
        )



from dataclasses import dataclass


@dataclass
class Point:
    """
    A geo-coordinate point with latitude and longitude.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    latitude: float
    longitude: float



from metamenth.energysystem.renewable_energy_system import RenewableEnergySystem
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import WindTurbineType


class WindMill(RenewableEnergySystem):
    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit,
                 number_of_turbine: int, turbine_type: WindTurbineType):
        super().__init__(name, inverter, unit)
        self._number_of_turbines = None
        self._turbine_type = None

        self.number_of_turbines = number_of_turbine
        self.turbine_type = turbine_type

    @property
    def number_of_turbines(self) -> int:
        return self._number_of_turbines

    @number_of_turbines.setter
    def number_of_turbines(self, value: int):
        if value is None:
            raise ValueError("number_of_turbines should be of type int")
        self._number_of_turbines = value

    @property
    def turbine_type(self) -> WindTurbineType:
        return self._turbine_type

    @turbine_type.setter
    def turbine_type(self, value: WindTurbineType):
        if value is None:
            raise ValueError("turbine_type should be of type WindTurbineType")
        self._turbine_type = value

    def __str__(self):
        return (
            f"WindMill("
            f"{super().__str__()}, "
            f"Number of Turbines: {self.number_of_turbines}, "
            f"Turbine Type: {self.turbine_type.value})"
        )



from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.structure.interfaces.abstract_floor_space import AbstractFloorSpace
from metamenth.enumerations import RoomType


class Room(AbstractFloorSpace):
    """
    Defines rooms on a floor of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(
        self,
        area: AbstractMeasure,
        name: str,
        room_type: RoomType,
        location: str = None
    ):
        """
        :param area: The area of the room.
        :param location: The location of the room (three words terminated with a period).
        :param name: The name of the room.
        :param room_type: The type of the room.
        """
        super().__init__(area, name, location)
        self._name = None
        self._room_type = None

        # call setters to apply validation
        self.name = name
        self.room_type = room_type

    @property
    def room_type(self) -> RoomType:
        return self._room_type

    @room_type.setter
    def room_type(self, value: RoomType):
        if value is not None:
            self._room_type = value
        else:
            raise ValueError("room_type must be of type RoomType")

    def accept(self, visitor):
        """
        visitor method to accept
        visit operation for the current room
        :param visitor: the visitor object
        """
        visitor.visit_room(self)

    def __str__(self):
        room_details = (
            f"Room ({super().__str__()} Room, "
            f"Room Type: {self.room_type})"
        )

        return room_details



import uuid
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.structure.material import Material
from metamenth.enumerations import LayerRoughness


class Layer:
    """
    A layer in the envelope of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(
        self,
        height: AbstractMeasure,
        length: AbstractMeasure,
        thickness: AbstractMeasure,
        material: Material,
        roughness: LayerRoughness,
        has_vapour_barrier: bool = False,
        has_air_barrier: bool = False,

    ):
        self._UID = str(uuid.uuid4())  # Generate a unique identifier
        self._height = None
        self._length = None
        self._thickness = None
        self._material = None
        self._roughness = None
        self._has_vapour_barrier = has_vapour_barrier
        self._has_air_barrier = has_air_barrier

        # apply validation with setters
        self.height = height
        self.length = length
        self.thickness = thickness
        self.material = material
        self.roughness = roughness

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def height(self) -> AbstractMeasure:
        return self._height

    @height.setter
    def height(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("height should be of type BinaryMeasure")
        self._height = value

    @property
    def length(self) -> AbstractMeasure:
        return self._length

    @length.setter
    def length(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("length should be of type BinaryMeasure")
        self._length = value

    @property
    def thickness(self) -> AbstractMeasure:
        return self._thickness

    @thickness.setter
    def thickness(self, value: AbstractMeasure):
        if value is None:
            raise ValueError("thickness should be of type BinaryMeasure")
        self._thickness = value

    @property
    def material(self) -> Material:
        return self._material

    @material.setter
    def material(self, value: Material):
        if value is None:
            raise ValueError("material cannot be None")
        self._material = value

    @property
    def roughness(self) -> LayerRoughness:
        return self._roughness

    @roughness.setter
    def roughness(self, value: LayerRoughness):
        if value is None:
            raise ValueError("roughness must be of MaterialRoughness type")
        self._roughness = value

    @property
    def has_vapour_barrier(self) -> bool:
        return self._has_vapour_barrier

    @has_vapour_barrier.setter
    def has_vapour_barrier(self, value: bool):
        self._has_vapour_barrier = value

    @property
    def has_air_barrier(self) -> bool:
        return self._has_air_barrier

    @has_air_barrier.setter
    def has_air_barrier(self, value: bool):
        self._has_air_barrier = value

    def __str__(self):
        material_str = f"Material: {str(self.material)}" if self.material else "Material: None"
        return (
            f"Layer("
            f"UID: {self.UID}, "
            f"Height: {self.height.value} {self.height.measurement_unit}, "
            f"Length: {self.length.value} {self.length.measurement_unit}, "
            f"Thickness: {self.thickness.value} {self.thickness.measurement_unit}, "
            f"Roughness: {self.roughness.value}, "
            f"Vapour Barrier: {self.has_vapour_barrier}, "
            f"Air Barrier: {self.has_air_barrier}, "
            f"{material_str})"
        )



from metamenth.datatypes.interfaces.abstract_range_measure import AbstractRangeMeasure
from metamenth.datatypes.measure import Measure


class ExponentialMeasure(AbstractRangeMeasure):
    """
    Exponential measurement

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(self, measure: Measure):
        super().__init__(measure)
        self.exponent = measure.exponent
        self.mantissa: measure.mantissa

    def __str__(self):
        return (
            f"ExponentialMeasure("
            f"Exponent: {self.exponent}, "
            f"Mantissa: {self.maximum}, "
            f"{super().__str__()})"
        )


from metamenth.energysystem.interfaces.abstract_energy_system import AbstractEnergySystem
from metamenth.energysystem.renewable_energy_system import RenewableEnergySystem
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import EnergySource
from metamenth.utils import EntityInsert
from metamenth.utils import EntityRemover
from metamenth.utils import StructureEntitySearch
from metamenth.enumerations import BuildingEntity
from metamenth.enumerations import BatteryTech
from metamenth.enumerations import CapacitorTech
from typing import Union


class ImmobileStorageEnergySystem(AbstractEnergySystem):
    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit, energy_source: EnergySource,
                 tech: Union[BatteryTech, CapacitorTech]):
        super().__init__(name, inverter, unit)
        self._energy_source = None
        self._technology = None
        self._renewable_sources: [RenewableEnergySystem] = []

        self.energy_source = energy_source
        self.technology = tech

    @property
    def energy_source(self) -> EnergySource:
        return self._energy_source

    @energy_source.setter
    def energy_source(self, value: EnergySource):
        if value is None:
            raise ValueError("energy_source should be of type EnergySource")
        self._energy_source = value

    @property
    def technology(self) -> Union[BatteryTech, CapacitorTech]:
        return self._technology

    @technology.setter
    def technology(self, value: Union[BatteryTech, CapacitorTech]):
        if value is None:
            raise ValueError("technology should be of type CapacitorTech or BatteryTech")
        self._technology = value

    def add_renewable_energy_source(self, renewable_energy_source: RenewableEnergySystem):
        """
        adds renewable energy source for storage system
        :param renewable_energy_source: The renewable energy source
        """
        EntityInsert.insert_building_entity(self._renewable_sources, renewable_energy_source,
                                            BuildingEntity.ENERGY_SYSTEM.value)

    def get_renewable_energy_source(self, name) -> RenewableEnergySystem:
        """
        Search renewable energy source by name
        :param name:  the name of the renewable energy source
        :return:
        """
        return StructureEntitySearch.search_by_name(self._renewable_sources, name)

    def remove_renewable_energy_source(self, renewable_energy_source: RenewableEnergySystem):
        """
        removes renewable energy source from storage system
        :param renewable_energy_source: The renewable energy source
        """
        EntityRemover.remove_building_entity(self._renewable_sources, renewable_energy_source)

    def __str__(self):
        return (
            f"{super().__str__()}, "
            f"Energy Source: {self.energy_source.value}, "
            f"Technology: {self.technology.value}, "
            f"Renewal Sources: {self._renewable_sources}"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class SensorLogType(AbstractEnum):
    """
    Describes how sensor values are recorded

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    POLLING = "Polling"
    CHANGE_OF_VALUE = "ChangeOfValue"



from metamenth.enumerations.abstract_enum import AbstractEnum


class BoilerCategory(AbstractEnum):
    """
    Types of heating

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    LPG = "LPG"
    NATURAL_GAS = "NaturalGas"
    OIL_FIRED = "OilFired"
    BIOMASS = "Biomass"
    WOOD_CHIPS = "WoodChips"
    ELECTRICAL = "Electrical"
    OTHER = "Other"




from metamenth.enumerations.abstract_enum import AbstractEnum


class DuctSubType(AbstractEnum):
    """
    Subtypes of ducts

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    FRESH_AIR = "FreshAir"
    RETURN_AIR = "ReturnAir"
    GLYCOL = "Glycol"
    HOT_WATER = "HotWater"
    COLD_WATER = "ColdWater"
    HOT_AND_COLD_WATER = "HotAndColdWater"
    OTHER = "Other"



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import PowerState
from metamenth.enumerations import CirculationPumpType


class CirculationPump(AbstractDuctConnectedComponent):
    def __init__(self, name: str, pump_type: CirculationPumpType, power_state: PowerState):
        """
        Models a circulation pump in a built environment
        :param name: the unique name of the heat exchanger
        :param pump_type: the type of chiller
        :param power_state: the power state of the chiller
        """
        super().__init__(name)
        self._pump_type = None
        self._power_state = None

        self.pump_type = pump_type
        self.power_state = power_state

    @property
    def pump_type(self) -> CirculationPumpType:
        return self._pump_type

    @pump_type.setter
    def pump_type(self, value: CirculationPumpType):
        if not value:
            raise ValueError("pump_type must be of type CirculationPumpType")
        self._pump_type = value

    @property
    def power_state(self) -> PowerState:
        return self._power_state

    @power_state.setter
    def power_state(self, value: PowerState):
        if value is not None:
            self._power_state = value
        else:
            raise ValueError("power_state must be of type PowerState")

    def __str__(self):
        return (
            f"Circulation Pump ({super().__str__()}"
            f"Pump Type: {self._pump_type.value}, "
            f"Power State: {self.power_state.value})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class DuctConnectionEntityType(AbstractEnum):
    """
    Duct connection object types

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    SOURCE = "Source"
    DESTINATION = "Destination"




from metamenth.energysystem.storage_system.immobile_storage_energy_system import ImmobileStorageEnergySystem
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import EnergySource
from metamenth.enumerations import CapacitorTech


class SuperCapacitor(ImmobileStorageEnergySystem):
    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit,
                 energy_source: EnergySource, tech: CapacitorTech):
        super().__init__(name, inverter, unit, energy_source, tech)

    def __str__(self):
        return (
            f"SuperCapacitor("
            f"{super().__str__()})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class WindTurbineType(AbstractEnum):
    """
    Types of Wind Turbine

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    VERTICAL_AXIS_WIND_TURBINE_ON_SHORE = "VAWTOnShore"
    HORIZONTAL_AXIS_WIND_TURBINE_ON_SHORE = "HAWTOnShore"
    VERTICAL_AXIS_WIND_TURBINE_OFF_SHORE = "VAWTOffShore"
    HORIZONTAL_AXIS_WIND_TURBINE_OFF_SHORE = "HAWTOffShore"
    OTHER = "Other"



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import AirVolumeType
from metamenth.subsystem.hvac_components.duct import Damper
from metamenth.datatypes.binary_measure import BinaryMeasure


class AirVolumeBox(AbstractDuctConnectedComponent):
    def __init__(self, name: str, air_volume_type: AirVolumeType,
                 has_heating_capability: bool = False, has_cooling_capability: bool = False):
        """
        Models an air volume box in a built environment
        :param name: the unique name of the air volume box
        :param has_heating_capability: indicates if the air volume box can heat air
        :param has_cooling_capability: indicates if the air volume box can cool air
        """
        super().__init__(name)
        self._air_volume_type = None
        self._inlet_dampers: [Damper] = []
        self._has_heating_capability = has_heating_capability
        self._has_cooling_capability = has_cooling_capability
        self._max_air_flow = None

        self.air_volume_type = air_volume_type

    @property
    def air_volume_type(self) -> AirVolumeType:
        return self._air_volume_type

    @air_volume_type.setter
    def air_volume_type(self, value: AirVolumeType):
        if not value:
            raise ValueError("air_volume_type must be of type AirVolumeType")
        self._air_volume_type = value

    @property
    def has_heating_capability(self) -> bool:
        return self._has_heating_capability

    @has_heating_capability.setter
    def has_heating_capability(self, value: bool):
       self._has_heating_capability = value

    @property
    def has_cooling_capability(self) -> bool:
        return self._has_cooling_capability

    @has_cooling_capability.setter
    def has_cooling_capability(self, value: bool):
        self._has_cooling_capability = value

    @property
    def inlet_dampers(self) -> [Damper]:
        return self._inlet_dampers

    @inlet_dampers.setter
    def inlet_dampers(self, value: [Damper]):
        self._inlet_dampers = value

    @property
    def max_air_flow(self) -> BinaryMeasure:
        return self._max_air_flow

    @max_air_flow.setter
    def max_air_flow(self, value: BinaryMeasure):
        self._max_air_flow = value

    def __str__(self):
        return (
            f"AirVolumeBox ({super().__str__()}"
            f"Air Volume Type: {self.air_volume_type.value}, "
            f"Inlet Dampers: {self.inlet_dampers}, "
            f"Maximum Air Flow: {self.max_air_flow}, "
            f"Has Cooling Capability: {self._has_cooling_capability}, "
            f"Has Heating Capability: {self.has_heating_capability})"
        )



from metamenth.subsystem.interfaces.abstract_subsystem import AbstractSubsystem
from metamenth.subsystem.hvac_system import HVACSystem


class BuildingControlSystem(AbstractSubsystem):

    def __init__(self, name: str):
        super().__init__(name)
        self._hvac_system = None

    @property
    def hvac_system(self) -> HVACSystem:
        return self._hvac_system

    @hvac_system.setter
    def hvac_system(self, value: HVACSystem):
        if value is not None:
            self._hvac_system = value
        else:
            raise ValueError("hvac_system must be of type HVACSystem")

    def __str__(self):
        return (
            f"BuildingControlSystem ({super().__str__()}"
            f"HVAC System: {self.hvac_system}"
        )




import uuid
from metamenth.structure.cover import Cover
from typing import List
from typing import Dict
from metamenth.utils import StructureSearch


class Envelope:
    """
    The envelope of a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(self):
        self._UID = str(uuid.uuid4())
        self._covers: List['Cover'] = []

    @property
    def UID(self) -> str:
        return self._UID

    def add_cover(self, cover: Cover):
        """
        Adds a cover to a building
        :param cover: the building cover e.g., wall, roof
        :return:
        """
        if cover is None:
            raise ValueError("cover must be of type Cover")
        self._covers.append(cover)

    def get_cover_by_uid(self, uid: str) -> Cover:
        """
        Retrieves a cover given the uid
        :param uid: the uid of the cover
        :return:
        """
        return StructureSearch.search_by_id(self._covers, uid)

    def get_covers(self, search_term: Dict = None) -> List[Cover]:
        """
        Retrieves covers given the attributes and their values
        :param search_term: the uid of the floor
        :return:
        """
        return StructureSearch.search(self._covers, search_term)

    def __str__(self):
        cover_details = "\n".join(str(cover) for cover in self._covers)
        return (
            f"Cover("
            f"UID: {self.UID}, "
            f"Layers:\n{cover_details})"
        )



from dataclasses import dataclass
from metamenth.enumerations import WaveForm
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure


@dataclass
class RatedDeviceMeasure:
    voltage_rating: AbstractMeasure
    current_rating: AbstractMeasure
    frequency: AbstractMeasure = None
    power_factor: float = 0.0
    phase: float = 0.0
    voltage_output: AbstractMeasure = None
    current_output: AbstractMeasure = None
    power_output: AbstractMeasure = None
    waveform: WaveForm = None
    efficiency: AbstractMeasure = None



from metamenth.energysystem.interfaces.abstract_electrical import AbstractElectrical
from metamenth.enumerations import PowerState
from metamenth.enumerations import ATSPowerSourceType
from metamenth.enumerations import ATSOperationMode
from metamenth.enumerations import ATSTransitionType
from metamenth.enumerations import ATSSwitchingMechanism


class AutomaticTransferSwitch(AbstractElectrical):
    def __init__(self, name: str, power_state: PowerState):
        super().__init__(name, power_state)
        self._power_source_type = None
        self._transition_type = None
        self._operation_mode = None
        self._switching_mechanism = None

    @property
    def power_source_type(self) -> ATSPowerSourceType:
        return self._power_source_type

    @power_source_type.setter
    def power_source_type(self, value: ATSPowerSourceType):
        self._power_source_type = value

    @property
    def transition_type(self) -> ATSTransitionType:
        return self._transition_type

    @transition_type.setter
    def transition_type(self, value: ATSTransitionType):
        self._transition_type = value

    @property
    def operation_mode(self) -> ATSOperationMode:
        return self._operation_mode

    @operation_mode.setter
    def operation_mode(self, value: ATSOperationMode):
        self._operation_mode = value

    @property
    def switching_mechanism(self) -> ATSSwitchingMechanism:
        return self._switching_mechanism

    @switching_mechanism.setter
    def switching_mechanism(self, value: ATSSwitchingMechanism):
        self._switching_mechanism = value

    def __str__(self):
        return (
            f"AutomaticTransferSwitch("
            f"{super().__str__()}"
            f"Power Source Type: {self.power_source_type}, "
            f"Transition Type: {self.transition_type}, "
            f"Operation Mode: {self.operation_mode}, "
            f"Switching Mechanism: {self.switching_mechanism})"
        )




from abc import ABC
from typing import Dict, Any
from metamenth.datatypes.continuous_measure import ContinuousMeasure
from uuid import uuid4
import sys
from metamenth.utils import StructureEntitySearch
from typing import List
from typing import Union
from metamenth.measure_instruments.sensor_data import SensorData
from metamenth.measure_instruments.trigger_history import TriggerHistory


class AbstractTransducer(ABC):
    def __init__(self,
                 name: str,
                 registry_id: str = None,
                 input_voltage_range: ContinuousMeasure = None,
                 input_current_range: ContinuousMeasure = None,
                 output_current_range: ContinuousMeasure = None,
                 output_voltage_range: ContinuousMeasure = None):
        """
        Describes a transducers (in a building)
        :param name: the unique name of the transducers
        :param registry_id: the registry id of the transducers
        :param input_voltage_range: the input voltage range of the transducers
        :param input_current_range: the input current range of the transducers
        :param output_current_range: the output current range of the transducers
        :param output_voltage_range: the output voltage range of the transducers
        """
        self.UID = uuid4()
        self.name = name
        self.input_voltage_range = input_voltage_range
        self.input_current_range = input_current_range
        self.registry_id = registry_id
        self.output_current_range = output_current_range
        self.output_voltage_range = output_voltage_range
        self.meta_data: Dict[str, Any] = {}
        self._data = []

    def add_data(self, data: Union[List[TriggerHistory], List[SensorData]]):
        if data is None:
            raise ValueError('data should be a list of SensorData or TriggerHistory')
        self._data.extend(data)

    def remove_data(self, data: Union[TriggerHistory, SensorData]):
        self._data.remove(data)

    def add_meta_data(self, key, value):
        """
        Adds metadata to transducers
        :param key: the key part of the metadata
        :param value: the value part of the metadata
        :return:
        """
        self.meta_data[key] = value

    def remove_meta_data(self, key):
        """
        removes metadata to transducers
        :param key: the key part of the metadata
        :return:
        """
        try:
            del self.meta_data[key]
        except KeyError as err:
            print(err, file=sys.stderr)

    def get_data(self, search_terms: Dict = None) -> Union[List[SensorData], List[TriggerHistory]]:
        """
        Search data by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [SensorData|TriggerHistory]:
        """
        return StructureEntitySearch.search(self._data, search_terms)

    def get_data_by_date(self, from_timestamp: str, to_timestamp: str = None) -> Union[List[SensorData],
                                                                                       List[TriggerHistory]]:
        """
        searches transducer data based on provided timestamp
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return: [SensorData|TriggerHistory]
        """
        return StructureEntitySearch.date_range_search(self._data, from_timestamp, to_timestamp)

    def get(self, attribute):
        return getattr(self, attribute, None)

    def __eq__(self, other):
        if isinstance(other, AbstractTransducer):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False

    def __str__(self):
        return (f"Unit: {self.UID}, Name: {self.name}, Registry ID: {self.registry_id}, "
                f"Input Voltage Range: {self.input_voltage_range}, "
                f"Output Voltage Range: {self.output_voltage_range}, "
                f"Input Current Range: {self.input_current_range}, "
                f"Output Current Range: {self.output_current_range}, "
                f"Metadata: {self.meta_data})")




from typing import List
from metamenth.structure.interfaces.abstract_space import AbstractSpace
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.utils import EntityRemover
from metamenth.utils import EntityInsert
from metamenth.measure_instruments.interfaces.abstract_reader import AbstractReader
from metamenth.utils import StructureEntitySearch
from typing import Dict
from metamenth.datatypes.interfaces.abstract_dynamic_entity import AbstractDynamicEntity
from metamenth.enumerations import BuildingEntity
from metamenth.subsystem.appliance import Appliance
from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.subsystem.interfaces.abstract_ventilation_component import AbstractVentilationComponent
from metamenth.transducers.interfaces.abstract_transducer import AbstractTransducer
from metamenth.transducers.sensor import Sensor
from metamenth.transducers.actuator import Actuator
from typing import Union
from metamenth.enumerations import SensorMeasure
from metamenth.energysystem.interfaces.abstract_common_energy_system import AbstractCommonEnergySystem
from metamenth.misc import Validate
from metamenth.subsystem.hvac_components.fan import Fan
from metamenth.subsystem.hvac_components.filter import Filter
from metamenth.subsystem.hvac_components.damper import Damper
from metamenth.subsystem.hvac_components.controller import Controller


class AbstractFloorSpace(AbstractSpace, AbstractDynamicEntity):
    """
    An abstract class for spaces on a floor

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, area: AbstractMeasure, name: str, location: str = None, meter: AbstractReader = None):
        """
        Models spaces on a building's floor
        :param area: the area of the space
        :param name: the name of the space
        :param location: the what3word location of the space
        """
        AbstractSpace.__init__(self, area, location)
        AbstractDynamicEntity.__init__(self)

        self._name = None
        self._adjacent_spaces: List[AbstractFloorSpace] = []
        self._meter = meter
        self._appliances: List[Appliance] = []
        self._hvac_components: Union[List[AbstractHVACComponent], List[AbstractVentilationComponent]] = []
        self._energy_systems: [AbstractCommonEnergySystem] = []
        # apply validation through setters
        self.name = name

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is not None:
            self._name = value
        else:
            raise ValueError("name must be a string")

    @property
    def meter(self) -> AbstractReader:
        return self._meter

    @meter.setter
    def meter(self, value: AbstractReader):
        if value:
            if value.meter_location != self.location:
                raise ValueError("what3words location of meter should be the same as space")
        self._meter = value

    def add_transducer(self, new_transducer: AbstractTransducer):
        """
        Adds sensors and/or actuators to entities (rooms, open spaces, equipment, etc.)
        :param new_transducer: a transducers to be added to this space
        :return:
        """
        if isinstance(new_transducer, Sensor):
            allowed_room_sensors = [SensorMeasure.OCCUPANCY, SensorMeasure.CARBON_DIOXIDE, SensorMeasure.DAYLIGHT,
                                    SensorMeasure.TEMPERATURE, SensorMeasure.HUMIDITY]
            if new_transducer.measure in allowed_room_sensors:
                super().add_transducer(new_transducer)
            else:
                raise ValueError(f'Space sensors must be one of the following: {allowed_room_sensors}')
        elif isinstance(new_transducer, Actuator):
            raise ValueError(f'Actuators cannot be added to rooms directly')

    def add_adjacent_space(self, space: 'AbstractFloorSpace'):
        """
        specifies (adds) which spaces (room and open spaces) are adjacent to other spaces
        :param space:
        :return:
        """
        EntityInsert.insert_building_entity(self._adjacent_spaces, space, BuildingEntity.ADJACENT_SPACE.value)

    def remove_adjacent_space(self, adjacent_space: 'AbstractFloorSpace'):
        """
        Removes adjacent space from a space (room and open space)
        :param adjacent_space: the adjacent space to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._adjacent_spaces, adjacent_space)

    def add_appliance(self, appliance: Appliance):
        """
        adds appliances to floor spaces
        :param appliance: the appliance to add
        :return:
        """
        EntityInsert.insert_building_entity(self._appliances, appliance, BuildingEntity.APPLIANCE.value)

    def remove_appliance(self, appliance: Appliance):
        """
        Removes appliance from a space (room and open space)
        :param appliance: the adjacent space to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._appliances, appliance)

    def add_hvac_component(self, hvac_component: Union[AbstractHVACComponent, AbstractVentilationComponent]):
        """
        adds hvac component to floor spaces
        :param hvac_component: the hvac component to add
        :return:
        """
        if Validate.is_hvac_component_allowed_in_space(hvac_component, [Fan, Damper, Filter, Controller], self):
            EntityInsert.insert_building_entity(self._hvac_components, hvac_component,
                                                BuildingEntity.HVAC_COMPONENT.value)

    def remove_hvac_component(self, hvac_component: Union[AbstractHVACComponent, AbstractVentilationComponent]):
        """
        Removes hvac component from a space (room and open space)
        :param hvac_component: the adjacent space to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._hvac_components, hvac_component)

    def get_hvac_components(self, search_terms: Dict = None) -> Union[List[AbstractHVACComponent],
    List[AbstractVentilationComponent]]:
        """
        Search appliances by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._hvac_components, search_terms)

    def get_adjacent_space_by_name(self, name) -> 'AbstractFloorSpace':
        """
        Search adjacent spaces by name
        :param name: the name of the adjacent space
        :return:
        """
        return StructureEntitySearch.search_by_name(self._adjacent_spaces, name)

    def get_adjacent_space_by_uid(self, uid) -> 'AbstractFloorSpace':
        """
        Search adjacent spaces by uid
        :param uid: the unique identifier of the adjacent spaces
        :return:
        """
        return StructureEntitySearch.search_by_id(self._adjacent_spaces, uid)

    def get_adjacent_spaces(self, search_terms: Dict = None) -> ['AbstractFloorSpace']:
        """
        Search adjacent spaces by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._adjacent_spaces, search_terms)

    def get_appliance_by_name(self, name) -> Appliance:
        """
        Search appliances by name
        :param name: the name of the appliance
        :return:
        """
        return StructureEntitySearch.search_by_name(self._appliances, name)

    def get_appliance_by_uid(self, uid) -> Appliance:
        """
        Search appliance by uid
        :param uid: the unique identifier of the appliance
        :return:
        """
        return StructureEntitySearch.search_by_id(self._appliances, uid)

    def get_appliances(self, search_terms: Dict = None) -> [Appliance]:
        """
        Search appliances by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._appliances, search_terms)

    def add_energy_system(self, energy_system: AbstractCommonEnergySystem):
        """
        adds energy system to floor spaces
        :param energy_system: the energy system to add
        :return:
        """
        EntityInsert.insert_building_entity(self._energy_systems, energy_system, BuildingEntity.HVAC_COMPONENT.value)

    def remove_energy_system(self, energy_system: AbstractCommonEnergySystem):
        """
        Removes energy system from a space (room and open space)
        :param energy_system: the energy system to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._energy_systems, energy_system)

    def get_energy_systems(self, search_terms: Dict = None) -> [AbstractCommonEnergySystem]:
        """
        Search energy systems by attribute values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._energy_systems, search_terms)

    def __eq__(self, other):
        # spaces on a floor are equal if they share the same name
        if isinstance(other, AbstractFloorSpace):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False

    def __str__(self) -> str:
        transducers_info = "\n".join([f" - Transducer: {transducer}" for transducer in self._transducers])
        appliances_info = "\n".join([f" - Appliance: {appliance}" for appliance in self._appliances])
        spaces_info = "\n".join([f" - Adjacent Space: {space}" for space in self._adjacent_spaces])
        hvac_component_info = "\n".join([f" - HVAC Components: {component}" for component in self._hvac_components])
        energy_system_info = "\n".join([f" - Energy Systems: {system}" for system in self._energy_systems])

        return (
            f"{super().__str__()}"
            f"Name: {self.name}, "
            f"Meter: {self.meter}, "
            f"Adjacent Spaces: {spaces_info}, "
            f"Transducers: {transducers_info}, "
            f"HVAC Components: {hvac_component_info}, "
            f"Energy Systems: {energy_system_info}, "
            f"Appliances: {appliances_info}\n"
        )



from enum import Enum


class SensorMeasureType(Enum):
    """
    Various sensor measurement types

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    THERMO_COUPLE_TYPE_A = "ThermoCoupleTypeA"
    THERMO_COUPLE_TYPE_B = "ThermoCoupleTypeB"
    THERMO_COUPLE_TYPE_C = "ThermoCoupleTypeC"
    THERMO_COUPLE_TYPE_D = "ThermoCoupleTypeD"
    THERMO_COUPLE_TYPE_E = "ThermoCoupleTypeE"
    THERMO_COUPLE_TYPE_F = "ThermoCoupleTypeF"
    PT_100 = "PT100"
    PIEZO_RESISTIVE_PRESSURE = "PiezoResistivePressure"
    ANALOGUE_PRESSURE = "AnaloguePressure"
    PASSIVE_INFRARED = "PassiveInfrared"



from metamenth.enumerations.abstract_enum import AbstractEnum


class EngineSubType(AbstractEnum):
    """
    Modes of Engine

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    DIESEL = "Diesel"
    BIO_DIESEL = "Biodiesel"
    NATURAL_GAS = "NaturalGas"
    HYDROGEN = "Hydrogen"
    PETROL = "Petrol"
    COAL = "Coal"
    PROPANE = "Propane"
    ETHANOL = "Ethanol"
    GEO_THERMAL = "GeoThermal"
    GASOLINE = "GasOnline"
    JETFUEL = "JetFuel"
    WOOD = "Wood"
    BIOMASS = "Biomass"
    CHARCOAL = "Charcoal"
    PEAT = "Peat"
    OIL = "Oil"
    NONE = "None"
    OTHER = "Other"



import sys
from metamenth.datatypes.observable_message import ObservableMessage


class StateTrackDecorator:
    """
    This decorator class wraps around methods that need their state
    to be logged whenever changed occur
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return lambda *args, **kwargs: self(instance, *args, **kwargs)

    def __call__(self, instance, *args, **kwargs):
        """

        :param instance: the instance of the class whose state needs to be logged
        :param args: arguments of method modifying instance state
        :param kwargs: key value argument of method modifying instance state
        :return:
        """
        try:
            if getattr(instance, 'track_state'):
                variable_name = self.func.__name__

                index = variable_name.find('_')
                # for methods such as add_room, add_open_space, remove "remove" and "add"
                if variable_name[:3] == 'add' or variable_name[:6] == 'remove':
                    variable_name = variable_name[index+1:]

                # check if after removing add and remove prefix, the resulting output is a class instance
                # variable. If not add s to the resulting variable
                if not hasattr(instance, variable_name) and variable_name[-1] != 's':
                    variable_name = variable_name + 's'

                instance.notify_observers(ObservableMessage(
                    instance.__class__.__name__,
                    instance.UID, {variable_name: getattr(instance, '_' + variable_name)}))
        except AttributeError as err:
            print(err, file=sys.stderr)
        return self.func(instance, *args, **kwargs)



from metamenth.enumerations.abstract_enum import AbstractEnum


class WaveForm(AbstractEnum):
    """
    Different types of waveforms

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    SINUSOIDAL = "Sinusoidal"
    SQUARE = "Square"
    MODIFIED_SINE = "ModifiedSine"
    OTHER = "Other"




from uuid import uuid4
from abc import ABC


class AbstractSubsystem(ABC):
    def __init__(self, name: str):
        """
        Defines parent class of all subsystems
        :param name:
        """
        self._UID = str(uuid4())
        self._name = None

        self.name = name

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is not None:
            self._name = value
        else:
            raise ValueError("name must be of type str")

    def __eq__(self, other):
        # subsystems are equal if they share the same name and UID
        if isinstance(other, AbstractSubsystem):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False

    def __str__(self):
        return (
            f"UID: {self.UID}, "
            f"Name: {self.name}"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class MeterMeasureMode(AbstractEnum):
    """
    Different types of floors in a building.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    MANUAL = "Manual"
    AUTOMATIC = "Automatic"



from metamenth.enumerations.abstract_enum import AbstractEnum


class SensorMeasure(AbstractEnum):
    """
    Various phenomena that can be measured by a sensor

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    TEMPERATURE = "Temperature"
    PRESSURE = "Pressure"
    GAS_VELOCITY = "GasVelocity"
    LIQUID_VELOCITY = "LiquidVelocity"
    LUMINANCE = "Luminance"
    CARBON_DIOXIDE = "CarbonDioxide"
    CARBON_MONOXIDE = "CarbonMonoxide"
    NOISE = "Noise"
    CURRENT = "Current"
    VOLTAGE = "Voltage"
    GLOBAL_RADIATION = "GlobalRadiation"
    DIRECT_RADIATION = "DirectRadiation"
    SMOKE = "Smoke"
    OCCUPANCY = "Occupancy"
    DAYLIGHT = "Daylight"
    AIR_VOLUME = "AirVolume"
    HUMIDITY = "Humidity"
    OTHER = "Other"



from metamenth.measure_instruments.interfaces.abstract_data_measure import AbstractDataMeasure
from metamenth.enumerations import OperationType
from metamenth.misc import Validate


class ElectricVehicleConnectivity(AbstractDataMeasure):
    """
    This class represents the reading values of EV charging meter.
    The unit of measurement depends on the phenomenon measured by a meter

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, value: float, start_time: str, end_time: str, operation_type: OperationType, vehicle_id: str):
        """
        :param value: The numerical value measured
        :param start_time: the start timestamp for charging/discharging
        :param end_time: the end timestamp of charging/discharging
        :param operation_type: either charging or discharging
        :param vehicle_id: the id of the vehicle to charged or discharged

        """
        super().__init__(value, start_time)
        self._end_time = None
        self._operation_type = None
        self._vehicle_uid = None

        self.end_time = end_time
        self.operation_type = operation_type
        self.vehicle_uid = vehicle_id

    @property
    def end_time(self) -> str:
        return self._end_time

    @end_time.setter
    def end_time(self, value: str):
        if value is None:
            raise ValueError('end_time must be of type str')
        self._end_time = Validate.parse_date(value)

    @property
    def operation_type(self) -> OperationType:
        return self._operation_type

    @operation_type.setter
    def operation_type(self, value: OperationType):
        if value is None:
            raise ValueError('operation_type must be of type OperationType')
        self._operation_type = value

    @property
    def vehicle_uid(self) -> str:
        return self._vehicle_uid

    @vehicle_uid.setter
    def vehicle_uid(self, value: str):
        if value is None:
            raise ValueError('vehicle_uid must be of type str')
        self._vehicle_uid = value




from metamenth.subsystem.interfaces.abstract_ventilation_component import AbstractVentilationComponent
from metamenth.enumerations import RadiantSlabType


class RadiantSlab(AbstractVentilationComponent):

    def __init__(self, name: str, radiate_slab_type: RadiantSlabType):
        super().__init__(name)
        self._radiant_slab_type = None

        self.radiant_slab_type = radiate_slab_type

    @property
    def radiant_slab_type(self) -> RadiantSlabType:
        return self._radiant_slab_type

    @radiant_slab_type.setter
    def radiant_slab_type(self, value: RadiantSlabType):
        if value is not None:
            self._radiant_slab_type = value
        else:
            raise ValueError("radiant_slab_type must be of type RadiantSlabType")

    def __str__(self):
        return (
            f"RadiantSlab ({super().__str__()}"
            f"Type: {self.radiant_slab_type})"
        )




from typing import Dict
import sys
from metamenth.enumerations.abstract_enum import AbstractEnum


class StructureSearch:
    """
    A visitor that search structure from a list of structures
    """

    def __init__(self):
        pass

    @staticmethod
    def search_by_id(structures, uid):
        """
        search structures by unique identifiers
        :param structures: the structures to search: floor, room, open spaces
        :param uid: the unique identifiers
        :return:
        """
        return StructureSearch.search_structure(structures, 'UID', uid)

    @staticmethod
    def search_by_number(structures, entity_number):
        """
        search structures by entity number
        :param structures: the structures to search: floor, room, open spaces
        :param entity_number: the number of the entity
        :return:
        """
        return StructureSearch.search_structure(structures, 'number', entity_number)

    @staticmethod
    def search_by_name(structures, name):
        """
        search structures by name
        :param structures: the structures to search: floor, room, open spaces
        :param name: name of the structure
        :return:
        """
        return StructureSearch.search_structure(structures, 'name', name)

    @staticmethod
    def search(structures, search_terms: Dict):
        """
        search structures based on attribute values
        :param structures: the structures to search: floor, room, open spaces
        :param search_terms: key value pair of attributes and their values
        :return:
        """
        from metamenth.structure.interfaces.abstract_space import AbstractSpace
        from metamenth.structure.layer import Layer
        from metamenth.structure.cover import Cover

        if search_terms is None:
            return structures

        results = []
        for structure in structures:
            if not isinstance(structure, AbstractSpace) and not isinstance(structure, Layer) and \
                not isinstance(structure, Cover):
                raise ValueError('{} is not a structure, layer or cover type'.format(structure))
            found = True
            try:
                for attribute, value in search_terms.items():
                    att_value = getattr(structure, attribute)
                    if isinstance(att_value, AbstractEnum):
                        att_value = att_value.value
                    if att_value != value:
                        found = False
                if found:
                    results.append(structure)
            except AttributeError as err:
                # TODO: log errors to file
                print(err, file=sys.stderr)

        return results

    @staticmethod
    def search_structure(structures, search_field, search_value):
        """
        Search for structure floors, rooms, open spaces in a building
        :param structures: the list of structures to search
        :param search_field: the search field of the structure to search
        :param search_value: the value to use in the search
        :return:
        """
        from metamenth.structure.interfaces.abstract_space import AbstractSpace
        from metamenth.structure.layer import Layer
        from metamenth.structure.cover import Cover
        for structure in structures:
            if not isinstance(structure, AbstractSpace) and not isinstance(structure, Layer) and \
                not isinstance(structure, Cover):
                raise ValueError('is not a structure, layer or cover type'.format(structure))
            try:
                if getattr(structure, search_field) == search_value:
                    return structure
            except AttributeError as err:
                # TODO: log errors to file
                print(err, file=sys.stderr)
        return None



from metamenth.visitors.interfaces.abstract_space_visitor import AbstractSpaceVisitor
from typing import Dict


class SpaceSearchVisitor(AbstractSpaceVisitor):
    """
    A concrete visitor that searches for sensors in
    building spaces or zones
    """

    def __init__(self, floor_criteria: Dict = None,
                 room_criteria: Dict = None, open_space_criteria: Dict = None, include_floor: bool = True):
        super().__init__(floor_criteria, room_criteria, open_space_criteria)
        self._include_floor = include_floor

    def visit_floor(self, floor):
        """
        override visit floor from AbstractSpace Visitor
        """
        if self._match_criteria(floor, self._floor_criteria):
            print(f'Visiting floor: {floor.number}')
            if self._include_floor:
                self.found_entities.append(floor)

            for room in floor.get_rooms():
                room.accept(self)

            for open_space in floor.get_open_spaces():
                open_space.accept(self)

    def visit_room(self, room):
        if self._match_criteria(room, self._room_criteria):
            self.found_entities.append(room)

    def visit_open_space(self, open_space):
        if self._match_criteria(open_space, self._open_space_criteria):
            self.found_entities.append(open_space)



from uuid import uuid4
from metamenth.datatypes.rated_device_measure import RatedDeviceMeasure
from typing import List
from typing import Dict
from metamenth.datatypes.continuous_measure import ContinuousMeasure
from metamenth.measure_instruments.meter import Meter
from metamenth.utils import StructureEntitySearch
from metamenth.utils import EntityRemover
from metamenth.utils import EntityInsert
from metamenth.datatypes.interfaces.abstract_dynamic_entity import AbstractDynamicEntity
from metamenth.enumerations import BuildingEntity
from metamenth.measure_instruments.status_measure import StatusMeasure
from metamenth.datatypes.schedulable_entity import SchedulableEntity


class AbstractHVACComponent(AbstractDynamicEntity):

    def __init__(self, name: str, meter: Meter = None, rated_device_measure: RatedDeviceMeasure = None):
        super().__init__()
        self._UID = str(uuid4())
        self._name = None
        self._meter = meter
        self._rated_device_measure = rated_device_measure
        self._schedulable_entity = SchedulableEntity()
        self._operating_conditions: List[ContinuousMeasure] = []
        self._spaces = []
        self._status_measure: [StatusMeasure] = []

        self.name = name

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is not None:
            self._name = value
        else:
            raise ValueError("name must be of type str")

    @property
    def meter(self) -> Meter:
        return self._meter

    @meter.setter
    def meter(self, value: Meter):
        self._meter = value

    @property
    def rated_device_measure(self) -> RatedDeviceMeasure:
        return self._rated_device_measure

    @rated_device_measure.setter
    def rated_device_measure(self, value: RatedDeviceMeasure):
        self._rated_device_measure = value

    @property
    def operating_conditions(self) -> [ContinuousMeasure]:
        return self._operating_conditions.copy() if self._operating_conditions else []

    @operating_conditions.setter
    def operating_conditions(self, value: [ContinuousMeasure]):
        if value is not None and type(value) is list:
            self._operating_conditions.extend(value)

    @property
    def schedulable_entity(self) -> SchedulableEntity:
        return self._schedulable_entity

    @schedulable_entity.setter
    def schedulable_entity(self, value: SchedulableEntity):
        if value is None:
            raise ValueError("schedules should be of type SchedulableEntity")
        self._schedulable_entity = value

    def add_spaces(self, space: []):
        """
        Adds spaces served by this HVAC component
        :param space: the space
        :return:
        """
        EntityInsert.insert_building_entity(self._spaces, space, BuildingEntity.FLOOR_SPACE.value)
        return self

    def remove_space(self, space):
        """
        Removes a space: floor, room, open space from a hvac component
        :param space: the space to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._spaces, space)

    def get_spaces(self, search_terms: Dict = None) -> []:
        """
        Search spaces served by this component by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._spaces, search_terms)

    def add_status_measure(self, status: StatusMeasure):
        """
        Adds status of hvac component schedule to this building
        :param status: the schedule
        :return:
        """
        EntityInsert.insert_building_entity(self._status_measure, status)
        return self

    def remove_status_measure(self, status):
        """
        Removes a status measure from a hvac component
        :param status: the status measure to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._status_measure, status)

    def get_status_measure(self, search_terms: Dict = None) -> [StatusMeasure]:
        """
        Search data by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [StatusMeasure]:
        """
        return StructureEntitySearch.search(self._status_measure, search_terms)

    def get_status_measure_by_date(self, from_timestamp: str, to_timestamp: str = None) -> [StatusMeasure]:
        """
        searches status data based on provided timestamp
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return: [StatusMeasure]
        """
        return StructureEntitySearch.date_range_search(self._status_measure, from_timestamp, to_timestamp)

    def __eq__(self, other):
        # subsystems are equal if they share the same name
        if isinstance(other, AbstractHVACComponent):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False

    def get(self, attribute):
        return getattr(self, attribute, None)

    def __str__(self):
        return (
            f"UID: {self.UID}, "
            f"Name: {self.name}, "
            f"Meter: {self.meter}, "
            f"Rated Device Measure: {self.rated_device_measure}, "
            f"Operational Schedule: {self._schedulable_entity}, "
            f"Operating Conditions: {self.operating_conditions}, "
        )


from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import BoilerCategory
from metamenth.enumerations import PowerState


class Boiler(AbstractDuctConnectedComponent):
    def __init__(self, name: str, category: BoilerCategory, power_state: PowerState):
        """
        Models a boiler in an hvac system
        :param name: the unique name of the boiler
        :param category: the boiler category
        :param power_state: the power state of the boiler
        :
        """
        super().__init__(name)
        self._category = None
        self._power_state = None

        self.power_state = power_state
        self.category = category

    @property
    def power_state(self) -> PowerState:
        return self._power_state

    @power_state.setter
    def power_state(self, value: PowerState):
        if value is not None:
            self._power_state = value
        else:
            raise ValueError("power_state must be of type PowerState")

    @property
    def category(self) -> BoilerCategory:
        return self._category

    @category.setter
    def category(self, value: BoilerCategory):
        if value is not None:
            self._category = value
        else:
            raise ValueError("category must be of type BoilerCategory")

    def __str__(self):
        return (
            f"Boiler ({super().__str__()}"
            f"Power State: {self.power_state}, "
            f"Category: {self.category})"
        )



from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from typing import Dict
from metamenth.subsystem.appliance import Appliance
from typing import Union
from metamenth.utils import StructureEntitySearch
from metamenth.utils import EntityRemover
from metamenth.utils import EntityInsert
from metamenth.enumerations import BuildingEntity
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.misc import Validate


class Controller(AbstractHVACComponent):
    def __init__(self, name: str):
        super().__init__(name)
        self._set_points: Dict[str, AbstractMeasure] = {}
        self._controller_entities: [Union[AbstractHVACComponent, Appliance]] = []

    def add_set_point(self, set_point: AbstractMeasure, transducer_pair: tuple):
        """
        Adds a set point for a controller
        :param set_point: the set point to be added
        :param transducer_pair: a tuple indicating the sensor and actuator this set point
        is being added. The formate is (sensor name, actuator name). Note that the sensor and actuator
        must exist before a set point can be added for them
        """
        if len(transducer_pair) != 2:
            raise ValueError('transducer_pair should be a tuple with the format (sensor_name, transducer_name)')

        controller_sensor = self.get_transducer_by_name(transducer_pair[0])
        controller_actuator = self.get_transducer_by_name(transducer_pair[1])
        if controller_sensor and controller_actuator:
            # validate the phenomenon measured by the sensor against the set point
            if set_point:
                if Validate.validate_sensor_type(controller_sensor.measure.value, set_point.measurement_unit.value):
                    self._set_points[f'{transducer_pair[0]}:{transducer_pair[1]}'] = set_point
                else:
                    raise ValueError('Sensor measure: {} not matching set point measure: {}'
                                     .format(controller_sensor.measure,
                                             set_point.measurement_unit))
        else:
            raise ValueError('There is no sensor/actuator found with the provided name for this controller')

    def get_set_point(self, sensor_name: str, actuator_name: str) -> AbstractMeasure:
        """
        Gets a set point
        :param sensor_name: the sensor associated with the set point
        :param actuator_name: the actuator associated with the set point
        """
        return self._set_points.get(f'{sensor_name}:{actuator_name}')

    def remove_set_point(self, sensor_name: str, actuator_name: str):
        """
        Removes a set point
        :param sensor_name: the sensor associated with the set point
        :param actuator_name: the actuator associated with the set point
        """
        if f'{sensor_name}:{actuator_name}' in self._set_points:
            del self._set_points[f'{sensor_name}:{actuator_name}']

    def add_controller_entity(self, entity: Union[AbstractHVACComponent, Appliance]):
        """
        Adds an entity controlled by this controller
        :param entity: the entity that is controlled
        """
        EntityInsert.insert_building_entity(self._controller_entities, entity, BuildingEntity.HVAC_COMPONENT.value)

    def remove_controller_entity(self, entity: Union[AbstractHVACComponent, Appliance]):
        """
        Removes an entity controlled by this controller
        :param entity: the entity to be removed
        """
        EntityRemover.remove_building_entity(self._controller_entities, entity)

    def get_controller_entities(self, search_terms: Dict = None) -> [Union[AbstractHVACComponent, Appliance]]:
        """
        Search data by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [Union[AbstractHVACComponent, AbstractDuctConnectedComponent, Appliance]]:
        """
        return StructureEntitySearch.search(self._controller_entities, search_terms)

    def __str__(self):
        return (
            f"Controller ({super().__str__()}"
            f"Set Points: {self._set_points}, "
            f"Controlled Entities: {self._controller_entities})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class ATSOperationMode(AbstractEnum):
    """
    ATS Operation modes

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    MANUAL = "Manual"
    NON_AUTOMATIC = "NonAutomatic"
    AUTOMATIC = "Automatic"



from metamenth.enumerations import BuildingEntity


class EntityInsert:
    """
    A visitor that remove entities from objects
    """

    def __init__(self):
        pass

    @staticmethod
    def insert_zone(entity, zone, building):
        # all unique zones are registered with the building
        if zone not in building.zones:
            building.zones.append(zone)
        if zone not in entity.zones:
            # add the space to the zone
            entity.zones.append(zone)
            from metamenth.structure.interfaces.abstract_space import AbstractSpace
            if isinstance(entity, AbstractSpace):
                zone.add_spaces([entity])

    @staticmethod
    def insert_building_entity(entity_list, entity, entity_type=None, entity_object=None):
        """
        Adds an entity to a space
        :param entity_list: the list of entity to add to (e.g., list of rooms or open space)
        :param entity: the entity (e.g. transducers, appliance, equipment, etc)
        :param entity_type: the type of entity
        :param entity_object: the entity object
        :return:
        """
        if entity_type in [BuildingEntity.TRANSDUCER.value, BuildingEntity.SCHEDULE.value,
                           BuildingEntity.ADJACENT_SPACE.value, BuildingEntity.APPLIANCE.value,
                           BuildingEntity.HVAC_COMPONENT.value, BuildingEntity.FLOOR.value]:
            # add transducer to room, open space or subsystem
            EntityInsert._insert_unique(entity_list, entity)

        elif entity_type == BuildingEntity.FLOOR_SPACE.value:
            # add open space or room to list of rooms

            for space in entity:
                EntityInsert._insert_unique(entity_list, space)

        elif entity_type == BuildingEntity.ZONE.value:
            if entity not in entity_list:
                # add the space to the zone
                entity.add_spaces([entity_object])
                entity_list.append(entity)

        elif entity_type in [BuildingEntity.OVERLAPPING_ZONE.value, BuildingEntity.ADJACENT_ZONE.value]:
            for new_zone in entity:
                # you can not add the same zone as an adjacent zone
                if entity_object == new_zone:
                    continue
                EntityInsert._insert_unique(entity_list, new_zone)
        else:
            entity_list.append(entity)

    @staticmethod
    def _insert_unique(entity_list, entity):
        if not entity in entity_list:
            entity_list.append(entity)



from dataclasses import dataclass
from typing import Dict
from datetime import datetime


@dataclass
class ObservableMessage:
    entity_type: str
    entity_id: str
    state: Dict
    message: str = None
    timestamp = datetime.now()



from dataclasses import dataclass
from metamenth.enumerations import MeasurementUnit


@dataclass
class Measure:
    unit: MeasurementUnit = None
    minimum: float = 0.0
    maximum: float = 0.0
    slope: float = 0.0
    exponent: float = 0.0
    mantissa: float = 0.0



from metamenth.enumerations.abstract_enum import AbstractEnum


class RadiantSlabType(AbstractEnum):
    """
    Types of radiant slabs

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    ELECTRIC = "Electric"
    AIR_HEATED = "AirHeated"
    HYDRONIC = "Hydronic"
    OTHER = "Other"




from metamenth.enumerations.abstract_enum import AbstractEnum


class EnergySource(AbstractEnum):
    """
    Types of Energy Sources

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    GRID = "Grid"
    Renewable = "Renewable"



from metamenth.enumerations.abstract_enum import AbstractEnum


class MeterAccumulationFrequency(AbstractEnum):
    """
    Data accumulation frequency of meters

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    DAILY = "Daily"
    WEEKLY = "Weekly"
    MONTHLY = "Monthly"
    QUARTERLY = "Quarterly"
    YEARLY = "Yearly"
    NONE = None




from metamenth.enumerations.abstract_enum import AbstractEnum


class ATSTransitionType(AbstractEnum):
    """
    ATS transition types

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    OPEN = "Open"
    CLOSE = "Close"



from metamenth.visitors.interfaces.abstract_space_visitor import AbstractSpaceVisitor
from typing import Dict
from metamenth.subsystem.hvac_components.duct import Duct
from metamenth.subsystem.hvac_components.air_volume_box import AirVolumeBox
import itertools


class HVACComponentSearchVisitor(AbstractSpaceVisitor):
    """
    A concrete visitor that searches for hvac component in
    building spaces
    """

    def __init__(self, hvac_component_criteria: Dict, floor_criteria: Dict = None,
                 room_criteria: Dict = None, open_space_criteria: Dict = None):
        """
        :param hvac_component_criteria: the search criteria for hvac components
        """
        super().__init__(floor_criteria, room_criteria, open_space_criteria)
        if 'component_class' not in hvac_component_criteria:
            raise ValueError(f'hvac component criteria must have component_class value: {hvac_component_criteria}')
        self._hvac_component_criteria = hvac_component_criteria

    def visit_room(self, room):
        if self._match_criteria(room, self._room_criteria):
            print(f'Visiting room: {room.name}')
            self._search_hvac_components(room)

    def visit_open_space(self, open_space):
        if self._match_criteria(open_space, self._open_space_criteria):
            print(f'Visiting open space: {open_space.name}')
            self._search_hvac_components(open_space)

    def _search_hvac_components(self, space):
        component_class = self._hvac_component_criteria['component_class']
        for hvac_component in space.get_hvac_components():
            # check if this hvac component is what we want to search
            if component_class == hvac_component.__class__.__name__:
                self._add_hvac_component(hvac_component)
            elif isinstance(hvac_component, Duct):
                # check if what we are looking for is inside a duct
                for duct_entity in itertools.chain(
                        hvac_component.get_heat_exchangers(),
                        hvac_component.get_fans(),
                        hvac_component.get_connected_air_volume_boxes(),
                        hvac_component.get_dampers(),
                        hvac_component.get_filters()
                ):
                    if component_class == duct_entity.__class__.__name__:
                        self._add_hvac_component(duct_entity)
            elif isinstance(hvac_component, AirVolumeBox):
                # check if what we are looking for is/are inlet dampers in air volume box
                if hvac_component.inlet_dampers:
                    if hvac_component.inlet_dampers[0].__class__.__name__ == component_class:
                        for damper in hvac_component.inlet_dampers:
                            self._add_hvac_component(damper)

    def _add_hvac_component(self, entity):
        if self._match_criteria(entity, self._hvac_component_criteria):
            self.found_entities.append(entity)



from metamenth.enumerations.measurement_unit import MeasurementUnit
from metamenth.enumerations.building_type import BuildingType
from metamenth.enumerations.floor_type import FloorType
from metamenth.enumerations.meter_type import MeterType
from metamenth.enumerations.zone_type import ZoneType
from metamenth.enumerations.open_space_type import OpenSpaceType
from metamenth.enumerations.room_type import RoomType
from metamenth.enumerations.data_measurement_type import DataMeasurementType
from metamenth.enumerations.hvac_type import HVACType
from metamenth.enumerations.material_type import MaterialType
from metamenth.enumerations.cover_type import CoverType
from metamenth.enumerations.building_entity import BuildingEntity
from metamenth.enumerations.trigger_type import TriggerType
from metamenth.enumerations.sensor_measure_type import SensorMeasureType
from metamenth.enumerations.sensor_measure import SensorMeasure
from metamenth.enumerations.recording_types import RecordingType
from metamenth.enumerations.sensor_log_type import SensorLogType
from metamenth.enumerations.meter_measure_mode import MeterMeasureMode
from metamenth.enumerations.meter_accumulation_frequency import MeterAccumulationFrequency
from metamenth.enumerations.appliacne_category import ApplianceCategory
from metamenth.enumerations.appliance_type import ApplianceType
from metamenth.enumerations.waveform import WaveForm
from metamenth.enumerations.duct_type import DuctType
from metamenth.enumerations.duct_sub_type import DuctSubType
from metamenth.enumerations.heat_exchanger_flow_type import HeatExchangerFlowType
from metamenth.enumerations.heat_exchanger_type import HeatExchangerType
from metamenth.enumerations.power_state import PowerState
from metamenth.enumerations.damper_type import DamperType
from metamenth.enumerations.duct_connection_entity_type import DuctConnectionEntityType
from metamenth.enumerations.ventilation_type import VentilationType
from metamenth.enumerations.radiant_slab_type import RadiantSlabType
from metamenth.enumerations.heating_type import HeatingType
from metamenth.enumerations.compressor_type import CompressorType
from metamenth.enumerations.boiler_category import BoilerCategory
from metamenth.enumerations.refrigerant_type import RefrigerantType
from metamenth.enumerations.coil_material import CoilMaterial
from metamenth.enumerations.heat_source import HeatSource
from metamenth.enumerations.chiller_type import ChillerType
from metamenth.enumerations.circulation_pump_type import CirculationPumpType
from metamenth.enumerations.air_volume_type import AirVolumeType
from metamenth.enumerations.cell_type import CellType
from metamenth.enumerations.solar_pv_type import SolarPVType
from metamenth.enumerations.wind_turbine_type import WindTurbineType
from metamenth.enumerations.engine_type import EngineType
from metamenth.enumerations.engine_sub_type import EngineSubType
from metamenth.enumerations.engine_mode import EngineMode
from metamenth.enumerations.energy_source import EnergySource
from metamenth.enumerations.capacitor_tech import CapacitorTech
from metamenth.enumerations.battery_tech import BatteryTech
from metamenth.enumerations.v2g_mode import V2GMode
from metamenth.enumerations.ats_transition_type import ATSTransitionType
from metamenth.enumerations.ats_power_source_type import ATSPowerSourceType
from metamenth.enumerations.ats_operation_mode import ATSOperationMode
from metamenth.enumerations.ats_switching_mechanism import ATSSwitchingMechanism
from metamenth.enumerations.ups_phase import UPSPhase
from metamenth.enumerations.operation_type import OperationType
from metamenth.enumerations.filter_type import FilterType
from metamenth.enumerations.layer_roughness import LayerRoughness
from metamenth.enumerations.pump_type import PumpType
from metamenth.enumerations.terrain_type import TerrainType
from metamenth.enumerations.solar_distribution_type import SolarDistributionType
from metamenth.enumerations.relationship_name import RelationshipName
from metamenth.enumerations.fcu_type import FCUType
from metamenth.enumerations.fcu_pipe_system import FCUPipeSystem



from metamenth.enumerations.abstract_enum import AbstractEnum


class MeasurementUnit(AbstractEnum):
    """
    Different types of measurement units

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    DEGREE_CELSIUS = "°C"
    PASCAL = "Pa"
    METERS_PER_SECOND = "m/s"
    R_VALUE = "R-Value"
    KILOWATTS = "kW"
    KILOWATTS_PER_HOUR = "kWh"
    DEGREES = "Deg"
    PERCENTAGE = "Percentage"
    WATTS_PER_METER_SQUARE = "W/m2"
    PRESENCE = "Presence"
    GRAM_PER_KILOGRAM = "g/kg"
    GRAM_PER_CUBIC_METER = "g/m3"
    LITERS_PER_SECOND = "l/s"
    CUBIC_METERS_PER_HOUR = "m3/h"
    CUBIC_FEET_PER_MINUTE = "ft3/min"
    MILLI_AMPERE = "mA"
    MICRO_AMPERE = "uA"
    AMPERE = "A"
    VOLT = "V"
    WATTS = "W"
    KILOGRAM_PER_CUBIC_METER = "kg/m3"
    POUNDS_PER_CUBIC_FOOT = "lb/ft3"
    WATTS_PER_SQUARE_METER_PER_DEGREE_CELSIUS = "w/(m2.°C)"
    CALORIES_PER_GRAM_PER_DEGREE_CELSIUS = "cal/g°C"
    JOULES_PER_KILOGRAM_PER_DEGREE_CELSIUS = "J/kg°C"
    KILOWATTS_PER_SQUARE_METER = "kW/m2"
    DECIBELS = "dB"
    PARTS_PER_MILLION = "ppm"
    LUX = "lx"
    CUBIC_METER_PER_SECOND = "m3/s"
    GALLONS_PER_SECOND = "gal/s"
    JOULES_PER_KELVIN = "J/K"
    WATTS_PER_SQUARE_METER_KELVIN = "W/(m2.K)"
    SQUARE_METERS_KELVIN_PER_WATTS = "m2.K/W"
    METERS = "m"
    CENTIMETERS = "cm"
    SQUARE_METERS = "m2"
    KILOGRAMS = "Kg"
    CUBIC_METER = "m3"
    CUBIC_CENTIMETER = "cm3"
    LITER = "l"
    CUBIC_FEET = "ft3"
    FEET_PER_SECOND = "ft/s"
    CANDELA_PER_SQUARE_METER = "cd/m2"
    NITS = "Nits (nt)"
    MICROGRAM_PER_CUBIC_METER = "µg/m3"
    SQUARE_FEET = "ft2"
    RELATIVE_HUMIDITY = "%RH"
    REVOLUTIONS_PER_MINUTE = "RMP"
    MEGAWATTS = "MW"







from metamenth.enumerations.abstract_enum import AbstractEnum


class HeatSource(AbstractEnum):
    """
    Heat source for heat pumps

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    AMBIENT_AIR = "AmbientAir"
    EXHAUST_AIR = "ExhaustAir"
    HORIZONTAL_GROUND_COLLECTOR = "HorizontalGroundCollector"
    VERTICAL_GROUND_COLLECTOR = "VerticalGroundCollector"
    WATER_WITH_ANTIFREEZE = "WaterWithAntiFreeze"
    AQUIFER = "Aquifer"
    OTHER = "Other"



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import HeatSource
from metamenth.subsystem.hvac_components.condenser import Condenser
from metamenth.subsystem.hvac_components.heat_exchanger import HeatExchanger
from metamenth.subsystem.hvac_components.compressor import Compressor


class HeatPump(AbstractDuctConnectedComponent):
    def __init__(self, name: str, heat_source: HeatSource,  condenser: Condenser = None, compressor: Compressor = None,
                 manufacturer: str = None):
        """
        Models a heat pump in a built environment
        :param name: the unique name of the heat exchanger
        :param heat_source: the type of heat exchanger
        :param manufacturer: the substance flow type of the heat exchanger
        """
        super().__init__(name)
        self._heat_source = None
        self._manufacturer = manufacturer
        self._condenser = condenser
        self._compressor = compressor
        self._heat_exchangers: [HeatExchanger] = []

        self.heat_source = heat_source

    @property
    def heat_source(self) -> HeatSource:
        return self._heat_source

    @heat_source.setter
    def heat_source(self, value: HeatSource):
        if not value:
            raise ValueError("heat_source must be of type HeatSource")
        self._heat_source = value

    @property
    def condenser(self) -> Condenser:
        return self._condenser

    @condenser.setter
    def condenser(self, value: Condenser):
        self._condenser = value

    @property
    def compressor(self) -> Compressor:
        return self._compressor

    @compressor.setter
    def compressor(self, value: Compressor):
        self._compressor = value

    @property
    def manufacturer(self) -> str:
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value: str):
        self._manufacturer = value

    @property
    def heat_exchangers(self) -> [HeatExchanger]:
        return self._heat_exchangers.copy()

    def add_heat_exchanger(self, heat_exchanger: HeatExchanger):
        if heat_exchanger:
            self._heat_exchangers.append(heat_exchanger)
        else:
            raise ValueError("heat_exchanger must be of type HeatExchanger")

    def __str__(self):
        return (
            f"HeatPump ({super().__str__()}"
            f"Heat Source: {self.heat_source.value}, "
            f"Condenser: {self.condenser}, "
            f"Compressor: {self.compressor}, "
            f"Heat Exchangers: {self._heat_exchangers}"
            f"Manufacturer : {self.manufacturer})"
        )



from enum import Enum


class BuildingEntity(Enum):
    """
    Various building Entity

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    BUILDING = "Building"
    FLOOR = "Floor"
    SCHEDULE = "Schedule"
    METER = "Meter"
    WEATHER_STATION = "WeatherStation"
    ROOM = "Room"
    OPEN_SPACE = "OpenSpace"
    TRANSDUCER = "Transducer"
    SPACE = "Space"
    ADJACENT_ZONE = "AdjacentZone"
    OVERLAPPING_ZONE = "OverlappingZone"
    ZONE = "Zone"
    ADJACENT_SPACE = "AdjacentSpace"
    SUBSYSTEM = "Subsystem"
    APPLIANCE = "Appliance"
    HVAC_COMPONENT = "HVACComponent"
    FLOOR_SPACE = "FloorSpace"
    ENERGY_SYSTEM = "EnergySystem"





from enum import Enum


class CoverType(Enum):
    """
    The different types of building cover

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    ROOF = "Roof"
    WALL = "Wall"
    FLOOR = "Floor"
    WINDOW = "Window"
    DOOR = "Door"



from enum import Enum


class HVACType(Enum):
    """
    Different types of HVAC (Thermal) Zones.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    PERIMETER = "Perimeter"
    INTERIOR = "Interior"
    NONE = None



import uuid
from metamenth.enumerations import CoverType
from metamenth.structure.layer import Layer
from typing import List
from metamenth.utils import StructureSearch
from typing import Dict


class Cover:
    """
       A building cover that forms the envelope of a building

       Author: Peter Yefi
       Email: peteryefi@gmail.com
       """
    def __init__(self, cover_type: CoverType):
        """
        :param cover_type: the type of building cover
        """
        self._UID = str(uuid.uuid4())
        self._cover_type = None
        self._layers: List['Layer'] = [] # the various layers in this building cover
        # validate cover type
        self.cover_type = cover_type

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def cover_type(self) -> CoverType:
        return self._cover_type

    @cover_type.setter
    def cover_type(self, value: CoverType):
        if value is None:
            raise ValueError("cover_type must be of type CoverType")
        self._cover_type = value

    def add_layer(self, layer: Layer):
        """
        Add a layer (e.g., external wall) to the building cover e.g., wall
        :param layer: the layer to be added
        :return:
        """
        if layer.material.material_type.value.split(":")[0].find(self.cover_type.value) != -1:
            self._layers.append(layer)
        else:
            raise ValueError("The layer you're trying to add has a different material from the cover.")

    def get_layer_by_uid(self, uid: str) -> Layer:
        """
        Retrieves a layer given the uid
        :param uid: the uid of the layer
        :return:
        """
        return StructureSearch.search_by_id(self._layers, uid)

    def get_layers(self, search_term: Dict = None) -> List[Layer]:
        """
        Retrieves layers given the attributes and their values
        :param search_term: the uid of the floor
        :return:
        """
        return StructureSearch.search(self._layers, search_term)

    def __str__(self):
        layer_str = "\n".join(str(layer) for layer in self._layers)
        return (
            f"Cover("
            f"UID: {self.UID}, "
            f"Cover Type: {self.cover_type}, "
            f"Layers:\n{layer_str})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class OpenSpaceType(AbstractEnum):
    """
    Open spaces on floors of buildings

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    PARKING_GARAGE = "Parking Garage"
    POOL_AREA = "Pool Area"
    OPEN_OFFICE = "Open Office"
    HALL = "Hall"
    CORRIDOR = "Corridor"
    STAIR_AREA = "StairArea"
    DINNING_AREA = "DinningArea"
    ESCALATOR_AREA = "EscalatorArea"
    RECEPTION_AREA = "ReceptionArea"
    TODDLERS_AREA = "ToddlersArea"
    OPEN_AREA = "OpenArea"
    SITTING_AREA = "SittingArea"
    OTHER = "Other"



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from metamenth.enumerations import CompressorType


class Compressor(AbstractDuctConnectedComponent):
    def __init__(self, name: str, compressor_type: CompressorType):
        """
        Models a compressor in hvac system
        :param name: the unique name of the compressor
        :param compressor_type: the type of compressor
        """
        super().__init__(name)
        self._compressor_type = None

        self.compressor_type = compressor_type

    @property
    def compressor_type(self) -> CompressorType:
        return self._compressor_type

    @compressor_type.setter
    def compressor_type(self, value: CompressorType):
        if value is not None:
            self._compressor_type = value
        else:
            raise ValueError("compressor_type must be of type CompressorType")

    def __str__(self):
        return (
            f"Compressor ({super().__str__()}"
            f"Type: {self.compressor_type})"
        )



from metamenth.energysystem.renewable_energy_system import RenewableEnergySystem
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import SolarPVType
from metamenth.enumerations import CellType
from metamenth.datatypes.binary_measure import BinaryMeasure


class SolarPV(RenewableEnergySystem):
    def __init__(self, name: str, inverter: bool, unit: MeasurementUnit,
                 solar_pv_type: SolarPVType, cell_type: CellType):
        super().__init__(name, inverter, unit)
        self._solar_pv_type = None
        self._cell_type = None
        self._thermal_capacity = None
        self._module_area = None

        self.cell_type = cell_type
        self.solar_pv_type = solar_pv_type

    @property
    def solar_pv_type(self) -> SolarPVType:
        return self._solar_pv_type

    @solar_pv_type.setter
    def solar_pv_type(self, value: SolarPVType):
        if value is None:
            raise ValueError("solar_pv_type should be of type SolarPVType")
        self._solar_pv_type = value

    @property
    def cell_type(self) -> CellType:
        return self._cell_type

    @cell_type.setter
    def cell_type(self, value: CellType):
        if value is None:
            raise ValueError("cell_type should be of type CellType")
        self._cell_type = value

    @property
    def thermal_capacity(self) -> BinaryMeasure:
        return self._thermal_capacity

    @thermal_capacity.setter
    def thermal_capacity(self, value: BinaryMeasure):
        self._thermal_capacity = value

    @property
    def module_area(self) -> BinaryMeasure:
        return self._module_area

    @module_area.setter
    def module_area(self, value: BinaryMeasure):
        self._module_area = value

    def __str__(self):
        return (
            f"SolarPV("
            f"{super().__str__()}, "
            f"Solar PV Type: {self.solar_pv_type.value}, "
            f"Cell Type: {self.cell_type.value}, "
            f"Thermal Capacity: {self.thermal_capacity}, "
            f"Module Area: {self.module_area})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class FCUType(AbstractEnum):
    """
    Types of fan coil units

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    CEILING_MOUNTED = "CeilingMounted"
    STANDALONE = "Standalone"





from enum import Enum


class ZoneType(Enum):
    """
    Different types of zones in a building.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    HVAC = "HVAC"
    LIGHTING = "Lighting"



from metamenth.enumerations.abstract_enum import AbstractEnum


class MeterType(AbstractEnum):
    """
    Different types of meters used in a building.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    ELECTRICITY = "Electricity"
    POWER = "Power"
    FLOW = "Flow"
    HEAT = "Heat"
    GAS = "Gas"




from abc import ABC
from metamenth.datatypes.measure import Measure
from metamenth.enumerations import DataMeasurementType


class AbstractMeasure(ABC):
    """
    Defines properties shared by all measures

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(self, measure: Measure, measure_type: DataMeasurementType = None):
        self.measurement_unit = measure.unit
        self.measure_type = measure_type

    def __str__(self):
        return f"Unit: {self.measurement_unit.value}, " \
               f"Measure Type: {self.measure_type.value if self.measure_type is not None else self.measure_type}"



from metamenth.measure_instruments.interfaces.abstract_data_measure import AbstractDataMeasure


class StatusMeasure(AbstractDataMeasure):

    def __init__(self, status: str, timestamp: str = None):
        """
        :param status: The string value measured
        :param timestamp: the time of measurement
        """
        super().__init__(status, timestamp)




from .entity_remover import EntityRemover
from .entity_insert import EntityInsert
from .search.structure_search import StructureSearch
from .search.structure_entity_search import StructureEntitySearch



from metamenth.datatypes.relationships.interfaces.abstract_relationship import AbstractRelationship
from metamenth.enumerations import RelationshipName
from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import AbstractDuctConnectedComponent
from typing import Union
from typing import List


class OneToMany(AbstractRelationship):

    def __init__(self, name: RelationshipName, hvac_component: Union[List[AbstractHVACComponent],
    List[AbstractDuctConnectedComponent]]):
        super().__init__(name)
        self.component = hvac_component



from metamenth.transducers.interfaces.abstract_transducer import AbstractTransducer
from typing import Optional
from metamenth.enumerations import SensorMeasure
from metamenth.enumerations import MeasurementUnit
from metamenth.enumerations import SensorMeasureType
from metamenth.misc import Validate
from metamenth.datatypes.interfaces.abstract_range_measure import AbstractRangeMeasure
from metamenth.enumerations import SensorLogType
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure


class Sensor(AbstractTransducer):

    def __init__(self, name: str, measure: SensorMeasure, unit: MeasurementUnit, measure_type: SensorMeasureType,
                 data_frequency: float, current_value: Optional[float] = None,
                 measure_range: AbstractRangeMeasure = None, sensor_log_type: SensorLogType = SensorLogType.POLLING):
        """
        :param name: the unique name of a sensor
        :param measure: the phenomenom (e.g., temperature) this sensor measures
        :param unit: the measurement unit of the data being measured
        :param measure_type: the type of data measured by the sensor
        :param data_frequency: what interval is the data recorded
        :param current_value: the current value for the sensor
        """
        super().__init__(name)
        self._measure = None
        self._data_frequency = None
        self._unit = None
        self._current_value = current_value
        self._measure_type = None
        self._measure_range = measure_range
        self._sensor_log_type = sensor_log_type

        # Setting values using setters to perform validation
        self.measure = measure
        self.data_frequency = data_frequency
        self.unit = unit
        self.measure_type = measure_type

        # validate sensor type and measurement
        if not Validate.validate_sensor_type(self.measure.value, self.unit.value):
            raise ValueError("{0} sensor can not have {1} measurement unit".format(measure.value, unit.value))

    @property
    def measure(self) -> SensorMeasure:
        return self._measure

    @measure.setter
    def measure(self, value: SensorMeasure):
        if value is not None:
            self._measure = value
        else:
            raise ValueError("measure must be of type SensorMeasure")

    @property
    def measure_range(self) -> AbstractRangeMeasure:
        return self._measure_range

    @measure_range.setter
    def measure_range(self, value: AbstractRangeMeasure):
        if value is not None:
            self._measure_range = value
        else:
            raise ValueError("measure must be of type AbstractRangeMeasure")

    @property
    def data_frequency(self) -> float:
        return self._data_frequency

    @data_frequency.setter
    def data_frequency(self, value: float):
        if value is not None:
            self._data_frequency = value
        else:
            raise ValueError("data_frequency must be float")

    @property
    def unit(self) -> MeasurementUnit:
        return self._unit

    @unit.setter
    def unit(self, value: MeasurementUnit):
        if value is not None:
            self._unit = value
        else:
            raise ValueError("unit must be of type MeasurementUnit")

    @property
    def current_value(self) -> float:
        return self._current_value

    @current_value.setter
    def current_value(self, value: float):
       self._current_value = value

    @property
    def measure_type(self) -> SensorMeasureType:
        return self._measure_type

    @measure_type.setter
    def measure_type(self, value: SensorMeasureType):
        if value is not None:
            self._measure_type = value
        else:
            raise ValueError("measure_type must be of type MeasureType")

    @property
    def sensor_log_type(self) -> SensorLogType:
        return self._sensor_log_type

    @sensor_log_type.setter
    def sensor_log_type(self, value: SensorLogType):
        if value is not None:
            self._sensor_log_type = value
        else:
            raise ValueError("measure must be of type SensorLogType")

    @property
    def set_point(self) -> AbstractMeasure:
        return self._set_point

    @set_point.setter
    def set_point(self, value: AbstractMeasure):
        self.set_set_point(value, self._unit)

    def __str__(self):
        sensor_data = "\n".join(str(data) for data in self._data)
        return (
            f"Sensor("
            f"{super().__str__()}, "
            f"UID: {self.UID}, "
            f"Name: {self.name}, "
            f"Measure: {self.measure}, "
            f"Measure Range: {self.measure_range}, "
            f"Data Frequency: {self.data_frequency}, "
            f"Unit: {self.unit}, "
            f"CurrentValue: {self.current_value}, "
            f"Measure Type: {self.measure_type}, "
            f"Log Type: {self.sensor_log_type.value}, "
            f"Data Count: {len(sensor_data)}\n"
            f"Data: {sensor_data})"
        )


from metamenth.measure_instruments.interfaces.abstract_data_measure import AbstractDataMeasure
from metamenth.enumerations import DataMeasurementType


class MeterMeasure(AbstractDataMeasure):
    """
    This class represents the reading values of a meter in a building.
    The unit of measurement depends on the phenomenon measured by a meter

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, value: float, timestamp: str = None, measurement_type: DataMeasurementType = None):
        """
        :param value: The numerical value measured
        :param timestamp: the time of measurement
        :param measurement_type: the type of the measurment, e.g., electricity consumption

        """
        super().__init__(value, timestamp, measurement_type)



from metamenth.datatypes.observable_message import ObservableMessage


class Observable:
    """
    parent class for all entities that need
    their state to be tracked overtime
    """
    def __init__(self):
        self._observers = []

    def add_observer(self, observer):
        self._observers.append(observer)

    def remove_observer(self, observer):
        self._observers.remove(observer)

    def notify_observers(self, message: ObservableMessage):
        for observer in self._observers:
            observer.log_state(message)



from metamenth.enumerations.abstract_enum import AbstractEnum


class LayerRoughness(AbstractEnum):
    """
    Different types of material roughness

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    VERY_ROUGH = "VeryRough (0.04m)"
    ROUGH = "Rough (0.01m)"
    MEDIUM_ROUGH = "MediumRough (0.005m)"
    SMOOTH = "Smooth (0.0005m)"
    VERY_SMOOTH = "VerySmooth (0.0001m)"




from metamenth.datatypes.binary_measure import BinaryMeasure
from metamenth.datatypes.linear_measure import LinearMeasure
from metamenth.datatypes.continuous_measure import ContinuousMeasure
from metamenth.datatypes.exponential_measure import ExponentialMeasure
from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.datatypes.measure import Measure


class MeasureFactory:
    """
    A factory class that returns various types of measures
    """
    @staticmethod
    def create_measure(measure_type: str, measure: Measure) -> AbstractMeasure:
        if measure_type == "Binary":
            return BinaryMeasure(measure)
        elif measure_type == "Linear":
            return LinearMeasure(measure)
        elif measure_type == "Continuous":
            return ContinuousMeasure(measure)
        elif measure_type == "Exponential":
            return ExponentialMeasure(measure)
        else:
            raise ValueError("Invalid measure type")



from enum import Enum
from fuzzywuzzy import fuzz


class AbstractEnum(Enum):

    @classmethod
    def get_enum_type(cls, value: str):
        try:
            value = value.replace(" ", "_").replace("-", "_").upper()
            closest_key = max(cls.__members__.keys(), key=lambda k: fuzz.ratio(value.lower(), k.lower()))
            return cls[closest_key]
        except KeyError:
            return None



from metamenth.enumerations.abstract_enum import AbstractEnum


class ATSSwitchingMechanism(AbstractEnum):
    """
    ATS Switching mechanisms

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    CONTRACTOR = "Contractor"
    MOLDED_CASE = "MoldedCase"
    POWER_FRAME = "PowerFrame"



from metamenth.datatypes.interfaces.abstract_measure import AbstractMeasure
from metamenth.datatypes.measure import Measure
from metamenth.enumerations import DataMeasurementType


class BinaryMeasure(AbstractMeasure):
    """
    Represents a binary measure with a value and a measurement unit.

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    def __init__(self, measure: Measure, weather_type: DataMeasurementType = None):
        super().__init__(measure, weather_type)
        self.value = measure.minimum

    def __str__(self):
        return (
            f"BinaryMeasure("
            f"Value: {self.value}, "
            f"{super().__str__()})"
        )



from metamenth.visitors.interfaces.abstract_space_visitor import AbstractSpaceVisitor
from typing import Dict


class SensorSearchVisitor(AbstractSpaceVisitor):
    """
    A concrete visitor that searches for sensors in
    building spaces or zones
    """
    def __init__(self, sensor_criteria: Dict, floor_criteria: Dict = None,
                 room_criteria: Dict = None, open_space_criteria: Dict = None):
        """
        :param sensor_criteria: the search criteria for sensors
        """
        super().__init__(floor_criteria, room_criteria, open_space_criteria)
        self._sensor_criteria = sensor_criteria

    def visit_room(self, room):
        if self._match_criteria(room, self._room_criteria):
            print(f'Visiting room: {room.name}')
            self._search_sensors(room)

    def visit_open_space(self, open_space):
        if self._match_criteria(open_space, self._open_space_criteria):
            print(f'Visiting open space: {open_space.name}')
            self._search_sensors(open_space)

    def _search_sensors(self, space):
        # search for space sensors
        for sensor in space.get_transducers():
            if self._match_criteria(sensor, self._sensor_criteria):
                self.found_entities.append(sensor)

        # search for HVAC component sensors
        self._search_entities(space.get_hvac_components())
        # search appliances for sensors
        self._search_entities(space.get_appliances())
        # search energy systems for sensors
        self._search_entities(space.get_energy_systems())

    def _search_entities(self, entities):
        for entity in entities:
            for sensor in entity.get_transducers():
                if self._match_criteria(sensor, self._sensor_criteria):
                    self.found_entities.append(sensor)




from abc import ABC
from metamenth.utils import EntityInsert
from metamenth.utils import EntityRemover
from metamenth.utils import StructureEntitySearch
from typing import Dict
from metamenth.enumerations import BuildingEntity


class AbstractZonalEntity(ABC):

    def __init__(self):
        """
        """
        self._zones = []

    @property
    def zones(self):
        return self._zones

    @zones.setter
    def zones(self, value):
        if value is not None:
            self._zones = value
        else:
            raise ValueError('zones must be of type [Zone]')

    def add_zone(self, zone, building):
        """
        Adds a zone to this floor
        :param zone: the zone
        :param building, the building which spaces requires a zone
        :return:
        """
        EntityInsert.insert_zone(self, zone, building)

    def remove_zone(self, zone):
        """
        Removes a zone from a space: floor, room, open space
        :param zone: the zone to be removed
        :return:
        """
        EntityRemover.remove_building_entity(self._zones, zone, BuildingEntity.ZONE.value, self)

    def get_zone_by_name(self, name):
        """
        Search zones by name
        :param name:  the name of the zone
        :return:
        """
        return StructureEntitySearch.search_by_name(self._zones, name)

    def get_zone_by_uid(self, uid):
        """
        Search zones by uid
        :param uid: the unique identifier of the overlapping zone
        :return:
        """
        return StructureEntitySearch.search_by_id(self._zones, uid)

    def get_zones(self, search_terms: Dict = None):
        """
        Search zones by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._zones, search_terms)


from metamenth.observers.interfaces.observer import Observer
from metamenth.datatypes.observable_message import ObservableMessage
from dataclasses import asdict


class StructureStateChangeLogger(Observer):
    """
    State logger for structure entities
    """

    def __init__(self):
        self._state_log: [ObservableMessage] = []

    def log_state(self, message: ObservableMessage):
        self._state_log.append(asdict(message))

    @property
    def state_log(self):
        return self._state_log



from datetime import datetime
import uuid
from abc import ABC
from metamenth.misc import Validate
from metamenth.enumerations import DataMeasurementType
from typing import Union


class AbstractDataMeasure(ABC):
    """
    This class represents the data recorded by sensors and meters
    The unit of measurement depends on the phenomenon measured by a meter or sensor

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, value: Union[float, str], timestamp: str = None, measurement_type: DataMeasurementType = None):
        """
        :param value: The numerical/string value measured
        :param timestamp: the time of measurement
        :param measurement_type: the type of the measurment, e.g., electricity consumption

        """
        self._UID = str(uuid.uuid4())
        self._timestamp = datetime.now().replace(microsecond=0) if timestamp is None else Validate.parse_date(timestamp)
        self._value = None
        self._measurement_type = measurement_type

        # Apply validation
        self.value = value

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float):
        if value is not None:
            self._value = value
        else:
            raise ValueError("Value must be a float")

    @property
    def measurement_type(self) -> DataMeasurementType:
        return self._measurement_type

    @measurement_type.setter
    def measurement_type(self, value: DataMeasurementType):
        self._measurement_type = value


    @property
    def timestamp(self) -> datetime:
        return self._timestamp

    def __str__(self):
        """
        :return: A formatted string of the meter readings.
        """
        return (f"DataMeasure (UID: {self.UID}, Value: {self.value}, "
                f"Timestamp: {self.timestamp}, Measurement Type: {self.measurement_type.value if self.measurement_type else None})")




from uuid import uuid4
from metamenth.measure_instruments.weather_data import WeatherData
from typing import List
from metamenth.misc import Validate
from typing import Dict
from metamenth.utils import StructureEntitySearch


class WeatherStation:
    def __init__(self, name: str, location: str = None):
        """
        :param location: The location of the weather station.
        """
        self._UID = str(uuid4())
        self._name = None
        self._location = Validate.validate_what3word(location)
        self._weather_data: List[WeatherData] = []

        self.name = name

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is not None:
            self._name = value
        else:
            ValueError('name must be of type str')

    @property
    def location(self) -> str:
        return self._location

    @location.setter
    def location(self, value: str):
        self._location = Validate.validate_what3word(value)

    def add_weather_data(self, weather_data: [WeatherData]):
        """
        Adds some data recordings to this WeatherStation.
        :param weather_data: some weather data recorded for the weather station.
        """
        self._weather_data.extend(weather_data)

    def get_weather_data(self, search_terms: Dict = None) -> [WeatherData]:
        """
        Search weather data by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return [WeatherData]:
        """
        return StructureEntitySearch.search(self._weather_data, search_terms)

    def get_weather_data_by_date(self, from_timestamp: str, to_timestamp: str = None) ->[WeatherData]:
        """
        searches weather data based on provided timestamp
        :param from_timestamp: the start timestamp
        :param to_timestamp: the end timestamp
        :return: [WeatherData]
        """
        return StructureEntitySearch.date_range_search(self._weather_data, from_timestamp, to_timestamp)

    def __eq__(self, other):
        # Weather stations are equal if they share the same name
        if isinstance(other, WeatherStation):
            # Check for equality based on the 'name' attribute
            return self.name == other.name
        return False

    def __str__(self):
        weather_station_details = (
            f"WeatherStation("
            f"UID: {self.UID}, "
            f"UID: {self.name}, "
            f"Location: {self.location}, "
            f"WeatherData Count: {len(self._weather_data)})"
        )
        weather_data = "\n".join(str(data) for data in self._weather_data)
        return f"{weather_station_details}\nWeather Data:\n{weather_data}"



from abc import ABC
from metamenth.transducers.interfaces.abstract_transducer import AbstractTransducer
from metamenth.subsystem.hvac_components.interfaces.abstract_hvac_component import AbstractHVACComponent
from metamenth.subsystem.hvac_components.controller import Controller
from typing import Union
from metamenth.subsystem.appliance import Appliance


class Actuator(AbstractTransducer, ABC):
    """
    A representation of an actuator in a building

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """

    def __init__(self, name: str, trigger_output: Union[AbstractHVACComponent, Appliance],
                 controller: Controller = None, actuation_interval: float = None):
        """
        :param name: the unique name of a transducers
        :param controller: the controller that triggers the actuator
        :param trigger_output: the device or equipment which is actuated
        """
        super().__init__(name)
        self._trigger_output = trigger_output
        self._controller = None
        self._actuation_interval = actuation_interval

        self.controller = controller

    @property
    def trigger_output(self) -> Union[AbstractHVACComponent, Appliance]:
        return self._trigger_output

    @trigger_output.setter
    def trigger_output(self, value: Union[AbstractHVACComponent, Appliance]):
        if value is None:
            raise ValueError('trigger_output must be of type AbstractHvACComponent or Appliance')
        self._trigger_output = value

    @property
    def actuation_interval(self) -> float:
        return self._actuation_interval

    @actuation_interval.setter
    def actuation_interval(self, value: float):
        self._actuation_interval = value

    @property
    def controller(self) -> Controller:
        return self._controller

    @controller.setter
    def controller(self, value: Controller):
        self._controller = value

    def __str__(self):
        trigger_data = "\n".join(str(data) for data in self._data)
        return (
            f"Sensor("
            f"{super().__str__()}, "
            f"UID: {self.UID}, "
            f"Name: {self.name}, "
            f"Controller: {self.controller}, "
            f"Trigger Output: {self.trigger_output}, "
            f"Trigger Value: {self.actuation_interval}, "
            f"Trigger Count: {len(trigger_data)}\n"
            f"Trigger History: {trigger_data})"
        )


from metamenth.enumerations.abstract_enum import AbstractEnum


class CellType(AbstractEnum):
    """
    Types of cells

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    MONOCRYSTALLINE = "MonoCrystalline"
    POLYCRYSTALLINE = "PolyCrystalline"
    AMORPHOUS = "Amorphous"
    OTHER = "Other"



from uuid import uuid4
from typing import List
from metamenth.subsystem.ventilation_system import VentilationSystem


class HVACSystem:

    def __init__(self):
        self._UID = str(uuid4())
        self._ventilation_systems: List[VentilationSystem] = []

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def ventilation_systems(self) -> [VentilationSystem]:
        return self._ventilation_systems.copy() if self._ventilation_systems else []

    def add_ventilation_system(self, ventilation_system:  VentilationSystem):
        if ventilation_system is not None:
            self._ventilation_systems.append(ventilation_system)
        else:
            raise ValueError("ventilation_system should be of type VentilationSystem")

    def __str__(self):
        return (
            f"HVACSystem("
            f"UID: {self.UID}, "
            f"Type: {self.ventilation_systems})"
        )


from metamenth.enumerations.abstract_enum import AbstractEnum


class OperationType(AbstractEnum):
    """
    Type of electric vehicle charging operation

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    CHARGING = "Charging"
    DISCHARGING = "Discharging"



from metamenth.enumerations.abstract_enum import AbstractEnum


class RelationshipName(AbstractEnum):
    CONTAINS = "Contains"
    USES = "Uses"
    FEEDS = "Feeds"
    PART_OF = "PartOf"
    CONNECTED_TO = "ConnectedTo"
    CONTROLS = "Controls"
    OPERATES_ON = "OperatesOn"
    MONITORS = "Monitors"
    LOCATED_IN = "LocatedIn"




from uuid import uuid4
from metamenth.enumerations import VentilationType
from metamenth.subsystem.hvac_components.duct import Duct
from metamenth.subsystem.interfaces.abstract_ventilation_component import AbstractVentilationComponent
from metamenth.utils import EntityRemover
from metamenth.utils import EntityInsert
from metamenth.utils import StructureEntitySearch
from typing import Dict
from typing import Union
from metamenth.energysystem.engine import Engine
from typing import List


class VentilationSystem:
    def __init__(self, ventilation_type: VentilationType, principal_duct: Duct):
        """
        Models ventilation system in a building
        :param ventilation_type: the type of ventilation
        :param principal_duct: the principal duct of the ventilation system
        """
        self._UID = str(uuid4())
        self._ventilation_type = None
        self._principal_duct = None
        self._components: [Union[AbstractVentilationComponent, Engine]] = []

        self.ventilation_type = ventilation_type
        self.principal_duct = principal_duct

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def ventilation_type(self) -> VentilationType:
        return self._ventilation_type

    @ventilation_type.setter
    def ventilation_type(self, value: VentilationType):
        if value is not None:
            self._ventilation_type = value
        else:
            raise ValueError("ventilation_type should be of type VentilationType")

    @property
    def principal_duct(self) -> Duct:
        return self._principal_duct

    @principal_duct.setter
    def principal_duct(self, value: Duct):
        if value is not None:
            self._principal_duct = value
        else:
            raise ValueError("principal_duct should be of type Duct")

    def get_components(self, search_terms: Dict = None) -> Union[List[AbstractVentilationComponent], List[Engine]]:
        """
        Search ventilation components by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._components, search_terms)

    def add_component(self, component: Union[AbstractVentilationComponent, Engine]):
        """
        Adds a ventilation component to the ventilation system
        :param component: the component to add
        :return:
        """
        EntityInsert.insert_building_entity(self._components, component)
        return self

    def remove_component(self, component: Union[AbstractVentilationComponent, Engine]):
        """
        Removes a ventilation component from the ventilation system
        :param component: the component to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._components, component)

    def __eq__(self, other):
        # ventilation systems are equal if they share the same UID
        if isinstance(other, Duct):
            # Check for equality based on the UID attribute
            return self.UID == other.UID
        return False

    def __str__(self):
        return (
            f"VentilationSystem("
            f"UID: {self.UID}, "
            f"Type: {self.ventilation_type}, "
            f"Principal Duct: {self.principal_duct}"
            f"Components: {self._components})"
        )


import uuid
from metamenth.datatypes.schedulable_entity import SchedulableEntity
from metamenth.datatypes.interfaces.abstract_dynamic_entity import AbstractDynamicEntity
from metamenth.measure_instruments.meter import Meter


class AbstractCommonEnergySystem(AbstractDynamicEntity):
    def __init__(self, name: str):
        super().__init__()
        self._UID = str(uuid.uuid4())
        self._name = None
        self._model = None
        self._manufacturer = None
        self._meter = None
        self._schedulable_entity = SchedulableEntity()

        self.name = name

    @property
    def UID(self) -> str:
        return self._UID

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if value is None:
            raise ValueError("name should be of type str")
        self._name = value

    @property
    def model(self) -> str:
        return self._model

    @model.setter
    def model(self, value: str):
        self._model = value

    @property
    def manufacturer(self) -> str:
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value: str):
        self._manufacturer = value

    @property
    def schedulable_entity(self) -> SchedulableEntity:
        return self._schedulable_entity

    @schedulable_entity.setter
    def schedulable_entity(self, value: SchedulableEntity):
        if value is None:
            raise ValueError("schedules should be of type SchedulableEntity")
        self._schedulable_entity = value

    @property
    def meter(self) -> Meter:
        return self._meter

    @meter.setter
    def meter(self, value: Meter):
        self._meter = value

    def __eq__(self, other):
        if isinstance(other, AbstractCommonEnergySystem):
            return self.name == other.name
        return False

    def __str__(self):
        return (
            f"UID: {self.UID}, "
            f"Name: {self.name}, "
            f"Manufacturer: {self.manufacturer}, "
            f"Manufacturing Year: {self.model}, "
            f"Meter: {self.meter}, "
            f"Operational Schedule: {self._schedulable_entity}, "
        )



from metamenth.subsystem.hvac_components.interfaces.abstract_duct_connected_component import (
    AbstractDuctConnectedComponent)
from metamenth.subsystem.hvac_components.heat_exchanger import HeatExchanger
from metamenth.subsystem.hvac_components.fan import Fan
from metamenth.enumerations import FCUType
from metamenth.enumerations import FCUPipeSystem


class FanCoilUnit(AbstractDuctConnectedComponent):
    def __init__(self, name: str, heat_exchanger: HeatExchanger,
                 fan: Fan, fcu_type: FCUType, fcu_pipe_system: FCUPipeSystem, is_ducted: bool = True):
        """
        Models a fan coil unit (FCU) in a built environment
        :param name: the unique name of the heat exchanger
        :param heat_exchanger: the heat exchanger which is part of the FCU
        :param fan: the fan which is part of the FCU
        :param fcu_type: the type of FCU
        :param fcu_pipe_system: the piping system of the FCU
        :param is_ducted: indicates if the FCU is connected to ventilation ducts or not
        """
        super().__init__(name)
        self._fan = None
        self._heat_exchanger = None
        self._fcu_type = None
        self._fcu_pipe_system = None
        self._is_ducted = is_ducted

        self.heat_exchanger = heat_exchanger
        self.fan = fan
        self.fcu_type = fcu_type
        self.fcu_pipe_system = fcu_pipe_system

    @property
    def heat_exchanger(self) -> HeatExchanger:
        return self._heat_exchanger

    @heat_exchanger.setter
    def heat_exchanger(self, value: HeatExchanger):
        if value is not None:
            self._heat_exchanger = value
        else:
            raise ValueError("heat_exchanger must be of type HeatExchanger")

    @property
    def fan(self) -> Fan:
        return self._fan

    @fan.setter
    def fan(self, value: Fan):
        if value is not None:
            self._fan = value
        else:
            raise ValueError("fan must be of type Fan")

    @property
    def fcu_type(self) -> FCUType:
        return self._fcu_type

    @fcu_type.setter
    def fcu_type(self, value: FCUType):
        if value is not None:
            self._fcu_type = value
        else:
            raise ValueError("fcu_type must be of type FCUType")

    @property
    def fcu_pipe_system(self) -> FCUPipeSystem:
        return self._fcu_pipe_system

    @fcu_pipe_system.setter
    def fcu_pipe_system(self, value: FCUPipeSystem):
        if value is not None:
            self._fcu_pipe_system = value
        else:
            raise ValueError("fcu_pipe_system must be of type FCUPipeSystem")

    @property
    def is_ducted(self) -> bool:
        return self._is_ducted

    @is_ducted.setter
    def is_ducted(self, value: bool):
        if value is not None:
            self._is_ducted = value
        else:
            raise ValueError("is_ducted must be of type bool")

    def __str__(self):
        return (
            f"FanCoilUnit ({super().__str__()}"
            f"FCU Type: {self.fcu_type}, "
            f"FCU Pipe System: {self.fcu_pipe_system}, "
            f"Is Ducted: {self.is_ducted}, "
            f"Heat Exchanger: {self.heat_exchanger}"
            f"Fan : {self.fan})"
        )



from metamenth.enumerations.abstract_enum import AbstractEnum


class DuctType(AbstractEnum):
    """
    Different types of ducts

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    AIR = "Air"
    WATER = "Water"
    WATER_WITH_ANTI_FREEZE = "WaterWithAntiFreeze"
    DHN = "DHN"
    OTHER = "Other"



from metamenth.enumerations.abstract_enum import AbstractEnum


class DamperType(AbstractEnum):
    """
    Types of dampers

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    BASIC_MANUAL_VOLUME = "BasicManualVolume"
    MANUAL_VOLUME = "ManualVolume"
    BACK_DRAFT = "BackDraft"
    BIOMETRIC_BYPASS = "BiometricBypass"
    OTHER = "Other"




from metamenth.enumerations.abstract_enum import AbstractEnum


class ChillerType(AbstractEnum):
    """
    Types of chillers

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    WATER_COOLED = "WaterCooled"
    AIR_COOLED = "AirCooled"
    OTHER = "Other"




from metamenth.enumerations import PowerState
from metamenth.energysystem.interfaces.abstract_common_energy_system import AbstractCommonEnergySystem


class AbstractElectrical(AbstractCommonEnergySystem):

    def __init__(self,  name: str, power_state: PowerState):
        super().__init__(name)
        self._power_state = None

        self.power_state = power_state

    @property
    def power_state(self) -> PowerState:
        return self._power_state

    @power_state.setter
    def power_state(self, value: PowerState):
        if value is None:
            raise ValueError('power_state must be on type PowerState')
        self._power_state = value

    def __str__(self):
        return (
            f"{super().__str__()}"
            f"Power State: {self.power_state}, "
        )


from abc import ABC
from metamenth.utils import EntityInsert
from metamenth.utils import EntityRemover
from metamenth.utils import StructureEntitySearch
from metamenth.transducers.interfaces.abstract_transducer import AbstractTransducer
from metamenth.enumerations import BuildingEntity
from typing import Dict


class AbstractDynamicEntity(ABC):

    def __init__(self):
        self._transducers: [AbstractTransducer] = []

    @property
    def transducers(self) -> [AbstractTransducer]:
        return self._transducers.copy()

    def add_transducer(self, new_transducer: AbstractTransducer):
        """
        Adds sensors and/or actuators to entities (rooms, open spaces, equipment, etc.)
        :param new_transducer: a transducers to be added to this space
        :return:
        """
        EntityInsert.insert_building_entity(self._transducers, new_transducer, BuildingEntity.TRANSDUCER.value)

    def remove_transducer(self, transducer: AbstractTransducer):
        """
        Removes a transducers from a subsystem
        :param transducer: the transducers to remove
        :return:
        """
        EntityRemover.remove_building_entity(self._transducers, transducer)

    def get_transducer_by_name(self, name) -> AbstractTransducer:
        """
        Search transducer by name
        :param name:  the name of the transducer
        :return:
        """
        return StructureEntitySearch.search_by_name(self._transducers, name)

    def get_transducer_by_uid(self, uid) -> AbstractTransducer:
        """
        Search transducers by uid
        :param uid: the unique identifier of the transducer
        :return:
        """
        return StructureEntitySearch.search_by_id(self._transducers, uid)

    def get_transducers(self, search_terms: Dict = None) -> ['AbstractTransducer']:
        """
        Search transducers by attributes values
        :param search_terms: a dictionary of attributes and their values
        :return:
        """
        return StructureEntitySearch.search(self._transducers, search_terms)


from metamenth.enumerations.abstract_enum import AbstractEnum


class ApplianceCategory(AbstractEnum):
    """
    Different categories of appliances

    Author: Peter Yefi
    Email: peteryefi@gmail.com
    """
    HOME = "Home"
    OFFICE = "Office"
    SMART = "Smart"
    OTHER = "Other"
    NONE = None

